# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime

# BreakpointId: Breakpoint identifier.
BreakpointId = str

# CallFrameId: Call frame identifier.
CallFrameId = str

# Location: Location in the source code.
class Location(ChromeTypeBase):
    def __init__(self,
                 scriptId: Union['Runtime.ScriptId'],
                 lineNumber: Union['int'],
                 columnNumber: Optional['int'] = None,
                 ):

        self.scriptId = scriptId
        self.lineNumber = lineNumber
        self.columnNumber = columnNumber


# ScriptPosition: Location in the source code.
class ScriptPosition(ChromeTypeBase):
    def __init__(self,
                 lineNumber: Union['int'],
                 columnNumber: Union['int'],
                 ):

        self.lineNumber = lineNumber
        self.columnNumber = columnNumber


# CallFrame: JavaScript call frame. Array of call frames form the call stack.
class CallFrame(ChromeTypeBase):
    def __init__(self,
                 callFrameId: Union['CallFrameId'],
                 functionName: Union['str'],
                 location: Union['Location'],
                 url: Union['str'],
                 scopeChain: Union['[Scope]'],
                 this: Union['Runtime.RemoteObject'],
                 functionLocation: Optional['Location'] = None,
                 returnValue: Optional['Runtime.RemoteObject'] = None,
                 ):

        self.callFrameId = callFrameId
        self.functionName = functionName
        self.functionLocation = functionLocation
        self.location = location
        self.url = url
        self.scopeChain = scopeChain
        self.this = this
        self.returnValue = returnValue


# Scope: Scope description.
class Scope(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 object: Union['Runtime.RemoteObject'],
                 name: Optional['str'] = None,
                 startLocation: Optional['Location'] = None,
                 endLocation: Optional['Location'] = None,
                 ):

        self.type = type
        self.object = object
        self.name = name
        self.startLocation = startLocation
        self.endLocation = endLocation


# SearchMatch: Search match for resource.
class SearchMatch(ChromeTypeBase):
    def __init__(self,
                 lineNumber: Union['float'],
                 lineContent: Union['str'],
                 ):

        self.lineNumber = lineNumber
        self.lineContent = lineContent


# BreakLocation: 
class BreakLocation(ChromeTypeBase):
    def __init__(self,
                 scriptId: Union['Runtime.ScriptId'],
                 lineNumber: Union['int'],
                 columnNumber: Optional['int'] = None,
                 type: Optional['str'] = None,
                 ):

        self.scriptId = scriptId
        self.lineNumber = lineNumber
        self.columnNumber = columnNumber
        self.type = type


class Debugger(PayloadMixin):
    """ Debugger domain exposes JavaScript debugging capabilities. It allows setting and removing
breakpoints, stepping through execution, exploring stack traces, etc.
    """
    @classmethod
    def continueToLocation(cls,
                           location: Union['Location'],
                           targetCallFrames: Optional['str'] = None,
                           ):
        """Continues execution until specific location is reached.
        :param location: Location to continue to.
        :type location: Location
        :param targetCallFrames: 
        :type targetCallFrames: str
        """
        return (
            cls.build_send_payload("continueToLocation", {
                "location": location,
                "targetCallFrames": targetCallFrames,
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Disables debugger for given page.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls,
               maxScriptsCacheSize: Optional['float'] = None,
               ):
        """Enables debugger for the given page. Clients should not assume that the debugging has been
enabled until the result for this command is received.
        :param maxScriptsCacheSize: The maximum size in bytes of collected scripts (not referenced by other heap objects)
the debugger can hold. Puts no limit if paramter is omitted.
        :type maxScriptsCacheSize: float
        """
        return (
            cls.build_send_payload("enable", {
                "maxScriptsCacheSize": maxScriptsCacheSize,
            }),
            cls.convert_payload({
                "debuggerId": {
                    "class": Runtime.UniqueDebuggerId,
                    "optional": False
                },
            })
        )

    @classmethod
    def evaluateOnCallFrame(cls,
                            callFrameId: Union['CallFrameId'],
                            expression: Union['str'],
                            objectGroup: Optional['str'] = None,
                            includeCommandLineAPI: Optional['bool'] = None,
                            silent: Optional['bool'] = None,
                            returnByValue: Optional['bool'] = None,
                            generatePreview: Optional['bool'] = None,
                            throwOnSideEffect: Optional['bool'] = None,
                            timeout: Optional['Runtime.TimeDelta'] = None,
                            ):
        """Evaluates expression on a given call frame.
        :param callFrameId: Call frame identifier to evaluate on.
        :type callFrameId: CallFrameId
        :param expression: Expression to evaluate.
        :type expression: str
        :param objectGroup: String object group name to put result into (allows rapid releasing resulting object handles
using `releaseObjectGroup`).
        :type objectGroup: str
        :param includeCommandLineAPI: Specifies whether command line API should be available to the evaluated expression, defaults
to false.
        :type includeCommandLineAPI: bool
        :param silent: In silent mode exceptions thrown during evaluation are not reported and do not pause
execution. Overrides `setPauseOnException` state.
        :type silent: bool
        :param returnByValue: Whether the result is expected to be a JSON object that should be sent by value.
        :type returnByValue: bool
        :param generatePreview: Whether preview should be generated for the result.
        :type generatePreview: bool
        :param throwOnSideEffect: Whether to throw an exception if side effect cannot be ruled out during evaluation.
        :type throwOnSideEffect: bool
        :param timeout: Terminate execution after timing out (number of milliseconds).
        :type timeout: Runtime.TimeDelta
        """
        return (
            cls.build_send_payload("evaluateOnCallFrame", {
                "callFrameId": callFrameId,
                "expression": expression,
                "objectGroup": objectGroup,
                "includeCommandLineAPI": includeCommandLineAPI,
                "silent": silent,
                "returnByValue": returnByValue,
                "generatePreview": generatePreview,
                "throwOnSideEffect": throwOnSideEffect,
                "timeout": timeout,
            }),
            cls.convert_payload({
                "result": {
                    "class": Runtime.RemoteObject,
                    "optional": False
                },
                "exceptionDetails": {
                    "class": Runtime.ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def getPossibleBreakpoints(cls,
                               start: Union['Location'],
                               end: Optional['Location'] = None,
                               restrictToFunction: Optional['bool'] = None,
                               ):
        """Returns possible locations for breakpoint. scriptId in start and end range locations should be
the same.
        :param start: Start of range to search possible breakpoint locations in.
        :type start: Location
        :param end: End of range to search possible breakpoint locations in (excluding). When not specified, end
of scripts is used as end of range.
        :type end: Location
        :param restrictToFunction: Only consider locations which are in the same (non-nested) function as start.
        :type restrictToFunction: bool
        """
        return (
            cls.build_send_payload("getPossibleBreakpoints", {
                "start": start,
                "end": end,
                "restrictToFunction": restrictToFunction,
            }),
            cls.convert_payload({
                "locations": {
                    "class": [BreakLocation],
                    "optional": False
                },
            })
        )

    @classmethod
    def getScriptSource(cls,
                        scriptId: Union['Runtime.ScriptId'],
                        ):
        """Returns source for the script with given id.
        :param scriptId: Id of the script to get source for.
        :type scriptId: Runtime.ScriptId
        """
        return (
            cls.build_send_payload("getScriptSource", {
                "scriptId": scriptId,
            }),
            cls.convert_payload({
                "scriptSource": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def getStackTrace(cls,
                      stackTraceId: Union['Runtime.StackTraceId'],
                      ):
        """Returns stack trace with given `stackTraceId`.
        :param stackTraceId: 
        :type stackTraceId: Runtime.StackTraceId
        """
        return (
            cls.build_send_payload("getStackTrace", {
                "stackTraceId": stackTraceId,
            }),
            cls.convert_payload({
                "stackTrace": {
                    "class": Runtime.StackTrace,
                    "optional": False
                },
            })
        )

    @classmethod
    def pause(cls):
        """Stops on the next JavaScript statement.
        """
        return (
            cls.build_send_payload("pause", {
            }),
            None
        )

    @classmethod
    def pauseOnAsyncCall(cls,
                         parentStackTraceId: Union['Runtime.StackTraceId'],
                         ):
        """
        :param parentStackTraceId: Debugger will pause when async call with given stack trace is started.
        :type parentStackTraceId: Runtime.StackTraceId
        """
        return (
            cls.build_send_payload("pauseOnAsyncCall", {
                "parentStackTraceId": parentStackTraceId,
            }),
            None
        )

    @classmethod
    def removeBreakpoint(cls,
                         breakpointId: Union['BreakpointId'],
                         ):
        """Removes JavaScript breakpoint.
        :param breakpointId: 
        :type breakpointId: BreakpointId
        """
        return (
            cls.build_send_payload("removeBreakpoint", {
                "breakpointId": breakpointId,
            }),
            None
        )

    @classmethod
    def restartFrame(cls,
                     callFrameId: Union['CallFrameId'],
                     ):
        """Restarts particular call frame from the beginning.
        :param callFrameId: Call frame identifier to evaluate on.
        :type callFrameId: CallFrameId
        """
        return (
            cls.build_send_payload("restartFrame", {
                "callFrameId": callFrameId,
            }),
            cls.convert_payload({
                "callFrames": {
                    "class": [CallFrame],
                    "optional": False
                },
                "asyncStackTrace": {
                    "class": Runtime.StackTrace,
                    "optional": True
                },
                "asyncStackTraceId": {
                    "class": Runtime.StackTraceId,
                    "optional": True
                },
            })
        )

    @classmethod
    def resume(cls):
        """Resumes JavaScript execution.
        """
        return (
            cls.build_send_payload("resume", {
            }),
            None
        )

    @classmethod
    def searchInContent(cls,
                        scriptId: Union['Runtime.ScriptId'],
                        query: Union['str'],
                        caseSensitive: Optional['bool'] = None,
                        isRegex: Optional['bool'] = None,
                        ):
        """Searches for given string in script content.
        :param scriptId: Id of the script to search in.
        :type scriptId: Runtime.ScriptId
        :param query: String to search for.
        :type query: str
        :param caseSensitive: If true, search is case sensitive.
        :type caseSensitive: bool
        :param isRegex: If true, treats string parameter as regex.
        :type isRegex: bool
        """
        return (
            cls.build_send_payload("searchInContent", {
                "scriptId": scriptId,
                "query": query,
                "caseSensitive": caseSensitive,
                "isRegex": isRegex,
            }),
            cls.convert_payload({
                "result": {
                    "class": [SearchMatch],
                    "optional": False
                },
            })
        )

    @classmethod
    def setAsyncCallStackDepth(cls,
                               maxDepth: Union['int'],
                               ):
        """Enables or disables async call stacks tracking.
        :param maxDepth: Maximum depth of async call stacks. Setting to `0` will effectively disable collecting async
call stacks (default).
        :type maxDepth: int
        """
        return (
            cls.build_send_payload("setAsyncCallStackDepth", {
                "maxDepth": maxDepth,
            }),
            None
        )

    @classmethod
    def setBlackboxPatterns(cls,
                            patterns: Union['[]'],
                            ):
        """Replace previous blackbox patterns with passed ones. Forces backend to skip stepping/pausing in
scripts with url matching one of the patterns. VM will try to leave blackboxed script by
performing 'step in' several times, finally resorting to 'step out' if unsuccessful.
        :param patterns: Array of regexps that will be used to check script url for blackbox state.
        :type patterns: []
        """
        return (
            cls.build_send_payload("setBlackboxPatterns", {
                "patterns": patterns,
            }),
            None
        )

    @classmethod
    def setBlackboxedRanges(cls,
                            scriptId: Union['Runtime.ScriptId'],
                            positions: Union['[ScriptPosition]'],
                            ):
        """Makes backend skip steps in the script in blackboxed ranges. VM will try leave blacklisted
scripts by performing 'step in' several times, finally resorting to 'step out' if unsuccessful.
Positions array contains positions where blackbox state is changed. First interval isn't
blackboxed. Array should be sorted.
        :param scriptId: Id of the script.
        :type scriptId: Runtime.ScriptId
        :param positions: 
        :type positions: [ScriptPosition]
        """
        return (
            cls.build_send_payload("setBlackboxedRanges", {
                "scriptId": scriptId,
                "positions": positions,
            }),
            None
        )

    @classmethod
    def setBreakpoint(cls,
                      location: Union['Location'],
                      condition: Optional['str'] = None,
                      ):
        """Sets JavaScript breakpoint at a given location.
        :param location: Location to set breakpoint in.
        :type location: Location
        :param condition: Expression to use as a breakpoint condition. When specified, debugger will only stop on the
breakpoint if this expression evaluates to true.
        :type condition: str
        """
        return (
            cls.build_send_payload("setBreakpoint", {
                "location": location,
                "condition": condition,
            }),
            cls.convert_payload({
                "breakpointId": {
                    "class": BreakpointId,
                    "optional": False
                },
                "actualLocation": {
                    "class": Location,
                    "optional": False
                },
            })
        )

    @classmethod
    def setBreakpointByUrl(cls,
                           lineNumber: Union['int'],
                           url: Optional['str'] = None,
                           urlRegex: Optional['str'] = None,
                           scriptHash: Optional['str'] = None,
                           columnNumber: Optional['int'] = None,
                           condition: Optional['str'] = None,
                           ):
        """Sets JavaScript breakpoint at given location specified either by URL or URL regex. Once this
command is issued, all existing parsed scripts will have breakpoints resolved and returned in
`locations` property. Further matching script parsing will result in subsequent
`breakpointResolved` events issued. This logical breakpoint will survive page reloads.
        :param lineNumber: Line number to set breakpoint at.
        :type lineNumber: int
        :param url: URL of the resources to set breakpoint on.
        :type url: str
        :param urlRegex: Regex pattern for the URLs of the resources to set breakpoints on. Either `url` or
`urlRegex` must be specified.
        :type urlRegex: str
        :param scriptHash: Script hash of the resources to set breakpoint on.
        :type scriptHash: str
        :param columnNumber: Offset in the line to set breakpoint at.
        :type columnNumber: int
        :param condition: Expression to use as a breakpoint condition. When specified, debugger will only stop on the
breakpoint if this expression evaluates to true.
        :type condition: str
        """
        return (
            cls.build_send_payload("setBreakpointByUrl", {
                "lineNumber": lineNumber,
                "url": url,
                "urlRegex": urlRegex,
                "scriptHash": scriptHash,
                "columnNumber": columnNumber,
                "condition": condition,
            }),
            cls.convert_payload({
                "breakpointId": {
                    "class": BreakpointId,
                    "optional": False
                },
                "locations": {
                    "class": [Location],
                    "optional": False
                },
            })
        )

    @classmethod
    def setBreakpointOnFunctionCall(cls,
                                    objectId: Union['Runtime.RemoteObjectId'],
                                    condition: Optional['str'] = None,
                                    ):
        """Sets JavaScript breakpoint before each call to the given function.
If another function was created from the same source as a given one,
calling it will also trigger the breakpoint.
        :param objectId: Function object id.
        :type objectId: Runtime.RemoteObjectId
        :param condition: Expression to use as a breakpoint condition. When specified, debugger will
stop on the breakpoint if this expression evaluates to true.
        :type condition: str
        """
        return (
            cls.build_send_payload("setBreakpointOnFunctionCall", {
                "objectId": objectId,
                "condition": condition,
            }),
            cls.convert_payload({
                "breakpointId": {
                    "class": BreakpointId,
                    "optional": False
                },
            })
        )

    @classmethod
    def setBreakpointsActive(cls,
                             active: Union['bool'],
                             ):
        """Activates / deactivates all breakpoints on the page.
        :param active: New value for breakpoints active state.
        :type active: bool
        """
        return (
            cls.build_send_payload("setBreakpointsActive", {
                "active": active,
            }),
            None
        )

    @classmethod
    def setPauseOnExceptions(cls,
                             state: Union['str'],
                             ):
        """Defines pause on exceptions state. Can be set to stop on all exceptions, uncaught exceptions or
no exceptions. Initial pause on exceptions state is `none`.
        :param state: Pause on exceptions mode.
        :type state: str
        """
        return (
            cls.build_send_payload("setPauseOnExceptions", {
                "state": state,
            }),
            None
        )

    @classmethod
    def setReturnValue(cls,
                       newValue: Union['Runtime.CallArgument'],
                       ):
        """Changes return value in top frame. Available only at return break position.
        :param newValue: New return value.
        :type newValue: Runtime.CallArgument
        """
        return (
            cls.build_send_payload("setReturnValue", {
                "newValue": newValue,
            }),
            None
        )

    @classmethod
    def setScriptSource(cls,
                        scriptId: Union['Runtime.ScriptId'],
                        scriptSource: Union['str'],
                        dryRun: Optional['bool'] = None,
                        ):
        """Edits JavaScript source live.
        :param scriptId: Id of the script to edit.
        :type scriptId: Runtime.ScriptId
        :param scriptSource: New content of the script.
        :type scriptSource: str
        :param dryRun: If true the change will not actually be applied. Dry run may be used to get result
description without actually modifying the code.
        :type dryRun: bool
        """
        return (
            cls.build_send_payload("setScriptSource", {
                "scriptId": scriptId,
                "scriptSource": scriptSource,
                "dryRun": dryRun,
            }),
            cls.convert_payload({
                "callFrames": {
                    "class": [CallFrame],
                    "optional": True
                },
                "stackChanged": {
                    "class": bool,
                    "optional": True
                },
                "asyncStackTrace": {
                    "class": Runtime.StackTrace,
                    "optional": True
                },
                "asyncStackTraceId": {
                    "class": Runtime.StackTraceId,
                    "optional": True
                },
                "exceptionDetails": {
                    "class": Runtime.ExceptionDetails,
                    "optional": True
                },
            })
        )

    @classmethod
    def setSkipAllPauses(cls,
                         skip: Union['bool'],
                         ):
        """Makes page not interrupt on any pauses (breakpoint, exception, dom exception etc).
        :param skip: New value for skip pauses state.
        :type skip: bool
        """
        return (
            cls.build_send_payload("setSkipAllPauses", {
                "skip": skip,
            }),
            None
        )

    @classmethod
    def setVariableValue(cls,
                         scopeNumber: Union['int'],
                         variableName: Union['str'],
                         newValue: Union['Runtime.CallArgument'],
                         callFrameId: Union['CallFrameId'],
                         ):
        """Changes value of variable in a callframe. Object-based scopes are not supported and must be
mutated manually.
        :param scopeNumber: 0-based number of scope as was listed in scope chain. Only 'local', 'closure' and 'catch'
scope types are allowed. Other scopes could be manipulated manually.
        :type scopeNumber: int
        :param variableName: Variable name.
        :type variableName: str
        :param newValue: New variable value.
        :type newValue: Runtime.CallArgument
        :param callFrameId: Id of callframe that holds variable.
        :type callFrameId: CallFrameId
        """
        return (
            cls.build_send_payload("setVariableValue", {
                "scopeNumber": scopeNumber,
                "variableName": variableName,
                "newValue": newValue,
                "callFrameId": callFrameId,
            }),
            None
        )

    @classmethod
    def stepInto(cls,
                 breakOnAsyncCall: Optional['bool'] = None,
                 ):
        """Steps into the function call.
        :param breakOnAsyncCall: Debugger will issue additional Debugger.paused notification if any async task is scheduled
before next pause.
        :type breakOnAsyncCall: bool
        """
        return (
            cls.build_send_payload("stepInto", {
                "breakOnAsyncCall": breakOnAsyncCall,
            }),
            None
        )

    @classmethod
    def stepOut(cls):
        """Steps out of the function call.
        """
        return (
            cls.build_send_payload("stepOut", {
            }),
            None
        )

    @classmethod
    def stepOver(cls):
        """Steps over the statement.
        """
        return (
            cls.build_send_payload("stepOver", {
            }),
            None
        )



class BreakpointResolvedEvent(BaseEvent):

    js_name = 'Debugger.breakpointResolved'
    hashable = ['breakpointId']
    is_hashable = True

    def __init__(self,
                 breakpointId: Union['BreakpointId', dict],
                 location: Union['Location', dict],
                 ):
        if isinstance(breakpointId, dict):
            breakpointId = BreakpointId(**breakpointId)
        self.breakpointId = breakpointId
        if isinstance(location, dict):
            location = Location(**location)
        self.location = location

    @classmethod
    def build_hash(cls, breakpointId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class PausedEvent(BaseEvent):

    js_name = 'Debugger.paused'
    hashable = ['asyncStackTraceId', 'asyncCallStackTraceId']
    is_hashable = True

    def __init__(self,
                 callFrames: Union['[CallFrame]', dict],
                 reason: Union['str', dict],
                 data: Union['dict', dict, None] = None,
                 hitBreakpoints: Union['[]', dict, None] = None,
                 asyncStackTrace: Union['Runtime.StackTrace', dict, None] = None,
                 asyncStackTraceId: Union['Runtime.StackTraceId', dict, None] = None,
                 asyncCallStackTraceId: Union['Runtime.StackTraceId', dict, None] = None,
                 ):
        if isinstance(callFrames, dict):
            callFrames = [CallFrame](**callFrames)
        self.callFrames = callFrames
        if isinstance(reason, dict):
            reason = str(**reason)
        self.reason = reason
        if isinstance(data, dict):
            data = dict(**data)
        self.data = data
        if isinstance(hitBreakpoints, dict):
            hitBreakpoints = [](**hitBreakpoints)
        self.hitBreakpoints = hitBreakpoints
        if isinstance(asyncStackTrace, dict):
            asyncStackTrace = Runtime.StackTrace(**asyncStackTrace)
        self.asyncStackTrace = asyncStackTrace
        if isinstance(asyncStackTraceId, dict):
            asyncStackTraceId = Runtime.StackTraceId(**asyncStackTraceId)
        self.asyncStackTraceId = asyncStackTraceId
        if isinstance(asyncCallStackTraceId, dict):
            asyncCallStackTraceId = Runtime.StackTraceId(**asyncCallStackTraceId)
        self.asyncCallStackTraceId = asyncCallStackTraceId

    @classmethod
    def build_hash(cls, asyncStackTraceId, asyncCallStackTraceId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ResumedEvent(BaseEvent):

    js_name = 'Debugger.resumed'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class ScriptFailedToParseEvent(BaseEvent):

    js_name = 'Debugger.scriptFailedToParse'
    hashable = ['scriptId', 'executionContextId']
    is_hashable = True

    def __init__(self,
                 scriptId: Union['Runtime.ScriptId', dict],
                 url: Union['str', dict],
                 startLine: Union['int', dict],
                 startColumn: Union['int', dict],
                 endLine: Union['int', dict],
                 endColumn: Union['int', dict],
                 executionContextId: Union['Runtime.ExecutionContextId', dict],
                 hash: Union['str', dict],
                 executionContextAuxData: Union['dict', dict, None] = None,
                 sourceMapURL: Union['str', dict, None] = None,
                 hasSourceURL: Union['bool', dict, None] = None,
                 isModule: Union['bool', dict, None] = None,
                 length: Union['int', dict, None] = None,
                 stackTrace: Union['Runtime.StackTrace', dict, None] = None,
                 ):
        if isinstance(scriptId, dict):
            scriptId = Runtime.ScriptId(**scriptId)
        self.scriptId = scriptId
        if isinstance(url, dict):
            url = str(**url)
        self.url = url
        if isinstance(startLine, dict):
            startLine = int(**startLine)
        self.startLine = startLine
        if isinstance(startColumn, dict):
            startColumn = int(**startColumn)
        self.startColumn = startColumn
        if isinstance(endLine, dict):
            endLine = int(**endLine)
        self.endLine = endLine
        if isinstance(endColumn, dict):
            endColumn = int(**endColumn)
        self.endColumn = endColumn
        if isinstance(executionContextId, dict):
            executionContextId = Runtime.ExecutionContextId(**executionContextId)
        self.executionContextId = executionContextId
        if isinstance(hash, dict):
            hash = str(**hash)
        self.hash = hash
        if isinstance(executionContextAuxData, dict):
            executionContextAuxData = dict(**executionContextAuxData)
        self.executionContextAuxData = executionContextAuxData
        if isinstance(sourceMapURL, dict):
            sourceMapURL = str(**sourceMapURL)
        self.sourceMapURL = sourceMapURL
        if isinstance(hasSourceURL, dict):
            hasSourceURL = bool(**hasSourceURL)
        self.hasSourceURL = hasSourceURL
        if isinstance(isModule, dict):
            isModule = bool(**isModule)
        self.isModule = isModule
        if isinstance(length, dict):
            length = int(**length)
        self.length = length
        if isinstance(stackTrace, dict):
            stackTrace = Runtime.StackTrace(**stackTrace)
        self.stackTrace = stackTrace

    @classmethod
    def build_hash(cls, scriptId, executionContextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ScriptParsedEvent(BaseEvent):

    js_name = 'Debugger.scriptParsed'
    hashable = ['scriptId', 'executionContextId']
    is_hashable = True

    def __init__(self,
                 scriptId: Union['Runtime.ScriptId', dict],
                 url: Union['str', dict],
                 startLine: Union['int', dict],
                 startColumn: Union['int', dict],
                 endLine: Union['int', dict],
                 endColumn: Union['int', dict],
                 executionContextId: Union['Runtime.ExecutionContextId', dict],
                 hash: Union['str', dict],
                 executionContextAuxData: Union['dict', dict, None] = None,
                 isLiveEdit: Union['bool', dict, None] = None,
                 sourceMapURL: Union['str', dict, None] = None,
                 hasSourceURL: Union['bool', dict, None] = None,
                 isModule: Union['bool', dict, None] = None,
                 length: Union['int', dict, None] = None,
                 stackTrace: Union['Runtime.StackTrace', dict, None] = None,
                 ):
        if isinstance(scriptId, dict):
            scriptId = Runtime.ScriptId(**scriptId)
        self.scriptId = scriptId
        if isinstance(url, dict):
            url = str(**url)
        self.url = url
        if isinstance(startLine, dict):
            startLine = int(**startLine)
        self.startLine = startLine
        if isinstance(startColumn, dict):
            startColumn = int(**startColumn)
        self.startColumn = startColumn
        if isinstance(endLine, dict):
            endLine = int(**endLine)
        self.endLine = endLine
        if isinstance(endColumn, dict):
            endColumn = int(**endColumn)
        self.endColumn = endColumn
        if isinstance(executionContextId, dict):
            executionContextId = Runtime.ExecutionContextId(**executionContextId)
        self.executionContextId = executionContextId
        if isinstance(hash, dict):
            hash = str(**hash)
        self.hash = hash
        if isinstance(executionContextAuxData, dict):
            executionContextAuxData = dict(**executionContextAuxData)
        self.executionContextAuxData = executionContextAuxData
        if isinstance(isLiveEdit, dict):
            isLiveEdit = bool(**isLiveEdit)
        self.isLiveEdit = isLiveEdit
        if isinstance(sourceMapURL, dict):
            sourceMapURL = str(**sourceMapURL)
        self.sourceMapURL = sourceMapURL
        if isinstance(hasSourceURL, dict):
            hasSourceURL = bool(**hasSourceURL)
        self.hasSourceURL = hasSourceURL
        if isinstance(isModule, dict):
            isModule = bool(**isModule)
        self.isModule = isModule
        if isinstance(length, dict):
            length = int(**length)
        self.length = length
        if isinstance(stackTrace, dict):
            stackTrace = Runtime.StackTrace(**stackTrace)
        self.stackTrace = stackTrace

    @classmethod
    def build_hash(cls, scriptId, executionContextId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime

# DatabaseWithObjectStores: Database with an array of object stores.
class DatabaseWithObjectStores(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 version: Union['float'],
                 objectStores: Union['[ObjectStore]'],
                 ):

        self.name = name
        self.version = version
        self.objectStores = objectStores


# ObjectStore: Object store.
class ObjectStore(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 keyPath: Union['KeyPath'],
                 autoIncrement: Union['bool'],
                 indexes: Union['[ObjectStoreIndex]'],
                 ):

        self.name = name
        self.keyPath = keyPath
        self.autoIncrement = autoIncrement
        self.indexes = indexes


# ObjectStoreIndex: Object store index.
class ObjectStoreIndex(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 keyPath: Union['KeyPath'],
                 unique: Union['bool'],
                 multiEntry: Union['bool'],
                 ):

        self.name = name
        self.keyPath = keyPath
        self.unique = unique
        self.multiEntry = multiEntry


# Key: Key.
class Key(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 number: Optional['float'] = None,
                 string: Optional['str'] = None,
                 date: Optional['float'] = None,
                 array: Optional['[Key]'] = None,
                 ):

        self.type = type
        self.number = number
        self.string = string
        self.date = date
        self.array = array


# KeyRange: Key range.
class KeyRange(ChromeTypeBase):
    def __init__(self,
                 lowerOpen: Union['bool'],
                 upperOpen: Union['bool'],
                 lower: Optional['Key'] = None,
                 upper: Optional['Key'] = None,
                 ):

        self.lower = lower
        self.upper = upper
        self.lowerOpen = lowerOpen
        self.upperOpen = upperOpen


# DataEntry: Data entry.
class DataEntry(ChromeTypeBase):
    def __init__(self,
                 key: Union['Runtime.RemoteObject'],
                 primaryKey: Union['Runtime.RemoteObject'],
                 value: Union['Runtime.RemoteObject'],
                 ):

        self.key = key
        self.primaryKey = primaryKey
        self.value = value


# KeyPath: Key path.
class KeyPath(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 string: Optional['str'] = None,
                 array: Optional['[]'] = None,
                 ):

        self.type = type
        self.string = string
        self.array = array


class IndexedDB(PayloadMixin):
    """ 
    """
    @classmethod
    def clearObjectStore(cls,
                         securityOrigin: Union['str'],
                         databaseName: Union['str'],
                         objectStoreName: Union['str'],
                         ):
        """Clears all entries from an object store.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        :param databaseName: Database name.
        :type databaseName: str
        :param objectStoreName: Object store name.
        :type objectStoreName: str
        """
        return (
            cls.build_send_payload("clearObjectStore", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
                "objectStoreName": objectStoreName,
            }),
            None
        )

    @classmethod
    def deleteDatabase(cls,
                       securityOrigin: Union['str'],
                       databaseName: Union['str'],
                       ):
        """Deletes a database.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        :param databaseName: Database name.
        :type databaseName: str
        """
        return (
            cls.build_send_payload("deleteDatabase", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
            }),
            None
        )

    @classmethod
    def deleteObjectStoreEntries(cls,
                                 securityOrigin: Union['str'],
                                 databaseName: Union['str'],
                                 objectStoreName: Union['str'],
                                 keyRange: Union['KeyRange'],
                                 ):
        """Delete a range of entries from an object store
        :param securityOrigin: 
        :type securityOrigin: str
        :param databaseName: 
        :type databaseName: str
        :param objectStoreName: 
        :type objectStoreName: str
        :param keyRange: Range of entry keys to delete
        :type keyRange: KeyRange
        """
        return (
            cls.build_send_payload("deleteObjectStoreEntries", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
                "objectStoreName": objectStoreName,
                "keyRange": keyRange,
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Disables events from backend.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables events from backend.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def requestData(cls,
                    securityOrigin: Union['str'],
                    databaseName: Union['str'],
                    objectStoreName: Union['str'],
                    indexName: Union['str'],
                    skipCount: Union['int'],
                    pageSize: Union['int'],
                    keyRange: Optional['KeyRange'] = None,
                    ):
        """Requests data from object store or index.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        :param databaseName: Database name.
        :type databaseName: str
        :param objectStoreName: Object store name.
        :type objectStoreName: str
        :param indexName: Index name, empty string for object store data requests.
        :type indexName: str
        :param skipCount: Number of records to skip.
        :type skipCount: int
        :param pageSize: Number of records to fetch.
        :type pageSize: int
        :param keyRange: Key range.
        :type keyRange: KeyRange
        """
        return (
            cls.build_send_payload("requestData", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
                "objectStoreName": objectStoreName,
                "indexName": indexName,
                "skipCount": skipCount,
                "pageSize": pageSize,
                "keyRange": keyRange,
            }),
            cls.convert_payload({
                "objectStoreDataEntries": {
                    "class": [DataEntry],
                    "optional": False
                },
                "hasMore": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def getMetadata(cls,
                    securityOrigin: Union['str'],
                    databaseName: Union['str'],
                    objectStoreName: Union['str'],
                    ):
        """Gets metadata of an object store
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        :param databaseName: Database name.
        :type databaseName: str
        :param objectStoreName: Object store name.
        :type objectStoreName: str
        """
        return (
            cls.build_send_payload("getMetadata", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
                "objectStoreName": objectStoreName,
            }),
            cls.convert_payload({
                "entriesCount": {
                    "class": float,
                    "optional": False
                },
                "keyGeneratorValue": {
                    "class": float,
                    "optional": False
                },
            })
        )

    @classmethod
    def requestDatabase(cls,
                        securityOrigin: Union['str'],
                        databaseName: Union['str'],
                        ):
        """Requests database with given name in given frame.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        :param databaseName: Database name.
        :type databaseName: str
        """
        return (
            cls.build_send_payload("requestDatabase", {
                "securityOrigin": securityOrigin,
                "databaseName": databaseName,
            }),
            cls.convert_payload({
                "databaseWithObjectStores": {
                    "class": DatabaseWithObjectStores,
                    "optional": False
                },
            })
        )

    @classmethod
    def requestDatabaseNames(cls,
                             securityOrigin: Union['str'],
                             ):
        """Requests database names for given security origin.
        :param securityOrigin: Security origin.
        :type securityOrigin: str
        """
        return (
            cls.build_send_payload("requestDatabaseNames", {
                "securityOrigin": securityOrigin,
            }),
            cls.convert_payload({
                "databaseNames": {
                    "class": [],
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

class DeviceOrientation(PayloadMixin):
    """ 
    """
    @classmethod
    def clearDeviceOrientationOverride(cls):
        """Clears the overridden Device Orientation.
        """
        return (
            cls.build_send_payload("clearDeviceOrientationOverride", {
            }),
            None
        )

    @classmethod
    def setDeviceOrientationOverride(cls,
                                     alpha: Union['float'],
                                     beta: Union['float'],
                                     gamma: Union['float'],
                                     ):
        """Overrides the Device Orientation.
        :param alpha: Mock alpha
        :type alpha: float
        :param beta: Mock beta
        :type beta: float
        :param gamma: Mock gamma
        :type gamma: float
        """
        return (
            cls.build_send_payload("setDeviceOrientationOverride", {
                "alpha": alpha,
                "beta": beta,
                "gamma": gamma,
            }),
            None
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

class Cast(PayloadMixin):
    """ A domain for interacting with Cast, Presentation API, and Remote Playback API
functionalities.
    """
    @classmethod
    def enable(cls,
               presentationUrl: Optional['str'] = None,
               ):
        """Starts observing for sinks that can be used for tab mirroring, and if set,
sinks compatible with |presentationUrl| as well. When sinks are found, a
|sinksUpdated| event is fired.
Also starts observing for issue messages. When an issue is added or removed,
an |issueUpdated| event is fired.
        :param presentationUrl: 
        :type presentationUrl: str
        """
        return (
            cls.build_send_payload("enable", {
                "presentationUrl": presentationUrl,
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Stops observing for sinks and issues.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def setSinkToUse(cls,
                     sinkName: Union['str'],
                     ):
        """Sets a sink to be used when the web page requests the browser to choose a
sink via Presentation API, Remote Playback API, or Cast SDK.
        :param sinkName: 
        :type sinkName: str
        """
        return (
            cls.build_send_payload("setSinkToUse", {
                "sinkName": sinkName,
            }),
            None
        )

    @classmethod
    def startTabMirroring(cls,
                          sinkName: Union['str'],
                          ):
        """Starts mirroring the tab to the sink.
        :param sinkName: 
        :type sinkName: str
        """
        return (
            cls.build_send_payload("startTabMirroring", {
                "sinkName": sinkName,
            }),
            None
        )

    @classmethod
    def stopCasting(cls,
                    sinkName: Union['str'],
                    ):
        """Stops the active Cast session on the sink.
        :param sinkName: 
        :type sinkName: str
        """
        return (
            cls.build_send_payload("stopCasting", {
                "sinkName": sinkName,
            }),
            None
        )



class SinksUpdatedEvent(BaseEvent):

    js_name = 'Cast.sinksUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 sinkNames: Union['[]', dict],
                 ):
        if isinstance(sinkNames, dict):
            sinkNames = [](**sinkNames)
        self.sinkNames = sinkNames

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class IssueUpdatedEvent(BaseEvent):

    js_name = 'Cast.issueUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 issueMessage: Union['str', dict],
                 ):
        if isinstance(issueMessage, dict):
            issueMessage = str(**issueMessage)
        self.issueMessage = issueMessage

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# TargetID: 
TargetID = str

# SessionID: Unique identifier of attached debugging session.
SessionID = str

# BrowserContextID: 
BrowserContextID = str

# TargetInfo: 
class TargetInfo(ChromeTypeBase):
    def __init__(self,
                 targetId: Union['TargetID'],
                 type: Union['str'],
                 title: Union['str'],
                 url: Union['str'],
                 attached: Union['bool'],
                 openerId: Optional['TargetID'] = None,
                 browserContextId: Optional['BrowserContextID'] = None,
                 ):

        self.targetId = targetId
        self.type = type
        self.title = title
        self.url = url
        self.attached = attached
        self.openerId = openerId
        self.browserContextId = browserContextId


# RemoteLocation: 
class RemoteLocation(ChromeTypeBase):
    def __init__(self,
                 host: Union['str'],
                 port: Union['int'],
                 ):

        self.host = host
        self.port = port


class Target(PayloadMixin):
    """ Supports additional targets discovery and allows to attach to them.
    """
    @classmethod
    def activateTarget(cls,
                       targetId: Union['TargetID'],
                       ):
        """Activates (focuses) the target.
        :param targetId: 
        :type targetId: TargetID
        """
        return (
            cls.build_send_payload("activateTarget", {
                "targetId": targetId,
            }),
            None
        )

    @classmethod
    def attachToTarget(cls,
                       targetId: Union['TargetID'],
                       flatten: Optional['bool'] = None,
                       ):
        """Attaches to the target with given id.
        :param targetId: 
        :type targetId: TargetID
        :param flatten: Enables "flat" access to the session via specifying sessionId attribute in the commands.
        :type flatten: bool
        """
        return (
            cls.build_send_payload("attachToTarget", {
                "targetId": targetId,
                "flatten": flatten,
            }),
            cls.convert_payload({
                "sessionId": {
                    "class": SessionID,
                    "optional": False
                },
            })
        )

    @classmethod
    def attachToBrowserTarget(cls):
        """Attaches to the browser target, only uses flat sessionId mode.
        """
        return (
            cls.build_send_payload("attachToBrowserTarget", {
            }),
            cls.convert_payload({
                "sessionId": {
                    "class": SessionID,
                    "optional": False
                },
            })
        )

    @classmethod
    def closeTarget(cls,
                    targetId: Union['TargetID'],
                    ):
        """Closes the target. If the target is a page that gets closed too.
        :param targetId: 
        :type targetId: TargetID
        """
        return (
            cls.build_send_payload("closeTarget", {
                "targetId": targetId,
            }),
            cls.convert_payload({
                "success": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def exposeDevToolsProtocol(cls,
                               targetId: Union['TargetID'],
                               bindingName: Optional['str'] = None,
                               ):
        """Inject object to the target's main frame that provides a communication
channel with browser target.

Injected object will be available as `window[bindingName]`.

The object has the follwing API:
- `binding.send(json)` - a method to send messages over the remote debugging protocol
- `binding.onmessage = json => handleMessage(json)` - a callback that will be called for the protocol notifications and command responses.
        :param targetId: 
        :type targetId: TargetID
        :param bindingName: Binding name, 'cdp' if not specified.
        :type bindingName: str
        """
        return (
            cls.build_send_payload("exposeDevToolsProtocol", {
                "targetId": targetId,
                "bindingName": bindingName,
            }),
            None
        )

    @classmethod
    def createBrowserContext(cls):
        """Creates a new empty BrowserContext. Similar to an incognito profile but you can have more than
one.
        """
        return (
            cls.build_send_payload("createBrowserContext", {
            }),
            cls.convert_payload({
                "browserContextId": {
                    "class": BrowserContextID,
                    "optional": False
                },
            })
        )

    @classmethod
    def getBrowserContexts(cls):
        """Returns all browser contexts created with `Target.createBrowserContext` method.
        """
        return (
            cls.build_send_payload("getBrowserContexts", {
            }),
            cls.convert_payload({
                "browserContextIds": {
                    "class": [BrowserContextID],
                    "optional": False
                },
            })
        )

    @classmethod
    def createTarget(cls,
                     url: Union['str'],
                     width: Optional['int'] = None,
                     height: Optional['int'] = None,
                     browserContextId: Optional['BrowserContextID'] = None,
                     enableBeginFrameControl: Optional['bool'] = None,
                     ):
        """Creates a new page.
        :param url: The initial URL the page will be navigated to.
        :type url: str
        :param width: Frame width in DIP (headless chrome only).
        :type width: int
        :param height: Frame height in DIP (headless chrome only).
        :type height: int
        :param browserContextId: The browser context to create the page in.
        :type browserContextId: BrowserContextID
        :param enableBeginFrameControl: Whether BeginFrames for this target will be controlled via DevTools (headless chrome only,
not supported on MacOS yet, false by default).
        :type enableBeginFrameControl: bool
        """
        return (
            cls.build_send_payload("createTarget", {
                "url": url,
                "width": width,
                "height": height,
                "browserContextId": browserContextId,
                "enableBeginFrameControl": enableBeginFrameControl,
            }),
            cls.convert_payload({
                "targetId": {
                    "class": TargetID,
                    "optional": False
                },
            })
        )

    @classmethod
    def detachFromTarget(cls,
                         sessionId: Optional['SessionID'] = None,
                         targetId: Optional['TargetID'] = None,
                         ):
        """Detaches session with given id.
        :param sessionId: Session to detach.
        :type sessionId: SessionID
        :param targetId: Deprecated.
        :type targetId: TargetID
        """
        return (
            cls.build_send_payload("detachFromTarget", {
                "sessionId": sessionId,
                "targetId": targetId,
            }),
            None
        )

    @classmethod
    def disposeBrowserContext(cls,
                              browserContextId: Union['BrowserContextID'],
                              ):
        """Deletes a BrowserContext. All the belonging pages will be closed without calling their
beforeunload hooks.
        :param browserContextId: 
        :type browserContextId: BrowserContextID
        """
        return (
            cls.build_send_payload("disposeBrowserContext", {
                "browserContextId": browserContextId,
            }),
            None
        )

    @classmethod
    def getTargetInfo(cls,
                      targetId: Optional['TargetID'] = None,
                      ):
        """Returns information about a target.
        :param targetId: 
        :type targetId: TargetID
        """
        return (
            cls.build_send_payload("getTargetInfo", {
                "targetId": targetId,
            }),
            cls.convert_payload({
                "targetInfo": {
                    "class": TargetInfo,
                    "optional": False
                },
            })
        )

    @classmethod
    def getTargets(cls):
        """Retrieves a list of available targets.
        """
        return (
            cls.build_send_payload("getTargets", {
            }),
            cls.convert_payload({
                "targetInfos": {
                    "class": [TargetInfo],
                    "optional": False
                },
            })
        )

    @classmethod
    def sendMessageToTarget(cls,
                            message: Union['str'],
                            sessionId: Optional['SessionID'] = None,
                            targetId: Optional['TargetID'] = None,
                            ):
        """Sends protocol message over session with given id.
        :param message: 
        :type message: str
        :param sessionId: Identifier of the session.
        :type sessionId: SessionID
        :param targetId: Deprecated.
        :type targetId: TargetID
        """
        return (
            cls.build_send_payload("sendMessageToTarget", {
                "message": message,
                "sessionId": sessionId,
                "targetId": targetId,
            }),
            None
        )

    @classmethod
    def setAutoAttach(cls,
                      autoAttach: Union['bool'],
                      waitForDebuggerOnStart: Union['bool'],
                      flatten: Optional['bool'] = None,
                      ):
        """Controls whether to automatically attach to new targets which are considered to be related to
this one. When turned on, attaches to all existing related targets as well. When turned off,
automatically detaches from all currently attached targets.
        :param autoAttach: Whether to auto-attach to related targets.
        :type autoAttach: bool
        :param waitForDebuggerOnStart: Whether to pause new targets when attaching to them. Use `Runtime.runIfWaitingForDebugger`
to run paused targets.
        :type waitForDebuggerOnStart: bool
        :param flatten: Enables "flat" access to the session via specifying sessionId attribute in the commands.
        :type flatten: bool
        """
        return (
            cls.build_send_payload("setAutoAttach", {
                "autoAttach": autoAttach,
                "waitForDebuggerOnStart": waitForDebuggerOnStart,
                "flatten": flatten,
            }),
            None
        )

    @classmethod
    def setDiscoverTargets(cls,
                           discover: Union['bool'],
                           ):
        """Controls whether to discover available targets and notify via
`targetCreated/targetInfoChanged/targetDestroyed` events.
        :param discover: Whether to discover available targets.
        :type discover: bool
        """
        return (
            cls.build_send_payload("setDiscoverTargets", {
                "discover": discover,
            }),
            None
        )

    @classmethod
    def setRemoteLocations(cls,
                           locations: Union['[RemoteLocation]'],
                           ):
        """Enables target discovery for the specified locations, when `setDiscoverTargets` was set to
`true`.
        :param locations: List of remote locations.
        :type locations: [RemoteLocation]
        """
        return (
            cls.build_send_payload("setRemoteLocations", {
                "locations": locations,
            }),
            None
        )



class AttachedToTargetEvent(BaseEvent):

    js_name = 'Target.attachedToTarget'
    hashable = ['sessionId']
    is_hashable = True

    def __init__(self,
                 sessionId: Union['SessionID', dict],
                 targetInfo: Union['TargetInfo', dict],
                 waitingForDebugger: Union['bool', dict],
                 ):
        if isinstance(sessionId, dict):
            sessionId = SessionID(**sessionId)
        self.sessionId = sessionId
        if isinstance(targetInfo, dict):
            targetInfo = TargetInfo(**targetInfo)
        self.targetInfo = targetInfo
        if isinstance(waitingForDebugger, dict):
            waitingForDebugger = bool(**waitingForDebugger)
        self.waitingForDebugger = waitingForDebugger

    @classmethod
    def build_hash(cls, sessionId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DetachedFromTargetEvent(BaseEvent):

    js_name = 'Target.detachedFromTarget'
    hashable = ['sessionId', 'targetId']
    is_hashable = True

    def __init__(self,
                 sessionId: Union['SessionID', dict],
                 targetId: Union['TargetID', dict, None] = None,
                 ):
        if isinstance(sessionId, dict):
            sessionId = SessionID(**sessionId)
        self.sessionId = sessionId
        if isinstance(targetId, dict):
            targetId = TargetID(**targetId)
        self.targetId = targetId

    @classmethod
    def build_hash(cls, sessionId, targetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ReceivedMessageFromTargetEvent(BaseEvent):

    js_name = 'Target.receivedMessageFromTarget'
    hashable = ['sessionId', 'targetId']
    is_hashable = True

    def __init__(self,
                 sessionId: Union['SessionID', dict],
                 message: Union['str', dict],
                 targetId: Union['TargetID', dict, None] = None,
                 ):
        if isinstance(sessionId, dict):
            sessionId = SessionID(**sessionId)
        self.sessionId = sessionId
        if isinstance(message, dict):
            message = str(**message)
        self.message = message
        if isinstance(targetId, dict):
            targetId = TargetID(**targetId)
        self.targetId = targetId

    @classmethod
    def build_hash(cls, sessionId, targetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class TargetCreatedEvent(BaseEvent):

    js_name = 'Target.targetCreated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 targetInfo: Union['TargetInfo', dict],
                 ):
        if isinstance(targetInfo, dict):
            targetInfo = TargetInfo(**targetInfo)
        self.targetInfo = targetInfo

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class TargetDestroyedEvent(BaseEvent):

    js_name = 'Target.targetDestroyed'
    hashable = ['targetId']
    is_hashable = True

    def __init__(self,
                 targetId: Union['TargetID', dict],
                 ):
        if isinstance(targetId, dict):
            targetId = TargetID(**targetId)
        self.targetId = targetId

    @classmethod
    def build_hash(cls, targetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class TargetCrashedEvent(BaseEvent):

    js_name = 'Target.targetCrashed'
    hashable = ['targetId']
    is_hashable = True

    def __init__(self,
                 targetId: Union['TargetID', dict],
                 status: Union['str', dict],
                 errorCode: Union['int', dict],
                 ):
        if isinstance(targetId, dict):
            targetId = TargetID(**targetId)
        self.targetId = targetId
        if isinstance(status, dict):
            status = str(**status)
        self.status = status
        if isinstance(errorCode, dict):
            errorCode = int(**errorCode)
        self.errorCode = errorCode

    @classmethod
    def build_hash(cls, targetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class TargetInfoChangedEvent(BaseEvent):

    js_name = 'Target.targetInfoChanged'
    hashable = []
    is_hashable = False

    def __init__(self,
                 targetInfo: Union['TargetInfo', dict],
                 ):
        if isinstance(targetInfo, dict):
            targetInfo = TargetInfo(**targetInfo)
        self.targetInfo = targetInfo

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# ApplicationCacheResource: Detailed application cache resource information.
class ApplicationCacheResource(ChromeTypeBase):
    def __init__(self,
                 url: Union['str'],
                 size: Union['int'],
                 type: Union['str'],
                 ):

        self.url = url
        self.size = size
        self.type = type


# ApplicationCache: Detailed application cache information.
class ApplicationCache(ChromeTypeBase):
    def __init__(self,
                 manifestURL: Union['str'],
                 size: Union['float'],
                 creationTime: Union['float'],
                 updateTime: Union['float'],
                 resources: Union['[ApplicationCacheResource]'],
                 ):

        self.manifestURL = manifestURL
        self.size = size
        self.creationTime = creationTime
        self.updateTime = updateTime
        self.resources = resources


# FrameWithManifest: Frame identifier - manifest URL pair.
class FrameWithManifest(ChromeTypeBase):
    def __init__(self,
                 frameId: Union['Page.FrameId'],
                 manifestURL: Union['str'],
                 status: Union['int'],
                 ):

        self.frameId = frameId
        self.manifestURL = manifestURL
        self.status = status


class ApplicationCache(PayloadMixin):
    """ 
    """
    @classmethod
    def enable(cls):
        """Enables application cache domain notifications.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getApplicationCacheForFrame(cls,
                                    frameId: Union['Page.FrameId'],
                                    ):
        """Returns relevant application cache data for the document in given frame.
        :param frameId: Identifier of the frame containing document whose application cache is retrieved.
        :type frameId: Page.FrameId
        """
        return (
            cls.build_send_payload("getApplicationCacheForFrame", {
                "frameId": frameId,
            }),
            cls.convert_payload({
                "applicationCache": {
                    "class": ApplicationCache,
                    "optional": False
                },
            })
        )

    @classmethod
    def getFramesWithManifests(cls):
        """Returns array of frame identifiers with manifest urls for each frame containing a document
associated with some application cache.
        """
        return (
            cls.build_send_payload("getFramesWithManifests", {
            }),
            cls.convert_payload({
                "frameIds": {
                    "class": [FrameWithManifest],
                    "optional": False
                },
            })
        )

    @classmethod
    def getManifestForFrame(cls,
                            frameId: Union['Page.FrameId'],
                            ):
        """Returns manifest URL for document in the given frame.
        :param frameId: Identifier of the frame containing document whose manifest is retrieved.
        :type frameId: Page.FrameId
        """
        return (
            cls.build_send_payload("getManifestForFrame", {
                "frameId": frameId,
            }),
            cls.convert_payload({
                "manifestURL": {
                    "class": str,
                    "optional": False
                },
            })
        )



class ApplicationCacheStatusUpdatedEvent(BaseEvent):

    js_name = 'Applicationcache.applicationCacheStatusUpdated'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['Page.FrameId', dict],
                 manifestURL: Union['str', dict],
                 status: Union['int', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = Page.FrameId(**frameId)
        self.frameId = frameId
        if isinstance(manifestURL, dict):
            manifestURL = str(**manifestURL)
        self.manifestURL = manifestURL
        if isinstance(status, dict):
            status = int(**status)
        self.status = status

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class NetworkStateUpdatedEvent(BaseEvent):

    js_name = 'Applicationcache.networkStateUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 isNowOnline: Union['bool', dict],
                 ):
        if isinstance(isNowOnline, dict):
            isNowOnline = bool(**isNowOnline)
        self.isNowOnline = isNowOnline

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# StorageId: DOM Storage identifier.
class StorageId(ChromeTypeBase):
    def __init__(self,
                 securityOrigin: Union['str'],
                 isLocalStorage: Union['bool'],
                 ):

        self.securityOrigin = securityOrigin
        self.isLocalStorage = isLocalStorage


# Item: DOM Storage item.
Item = [str]

class DOMStorage(PayloadMixin):
    """ Query and modify DOM storage.
    """
    @classmethod
    def clear(cls,
              storageId: Union['StorageId'],
              ):
        """
        :param storageId: 
        :type storageId: StorageId
        """
        return (
            cls.build_send_payload("clear", {
                "storageId": storageId,
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Disables storage tracking, prevents storage events from being sent to the client.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables storage tracking, storage events will now be delivered to the client.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getDOMStorageItems(cls,
                           storageId: Union['StorageId'],
                           ):
        """
        :param storageId: 
        :type storageId: StorageId
        """
        return (
            cls.build_send_payload("getDOMStorageItems", {
                "storageId": storageId,
            }),
            cls.convert_payload({
                "entries": {
                    "class": [Item],
                    "optional": False
                },
            })
        )

    @classmethod
    def removeDOMStorageItem(cls,
                             storageId: Union['StorageId'],
                             key: Union['str'],
                             ):
        """
        :param storageId: 
        :type storageId: StorageId
        :param key: 
        :type key: str
        """
        return (
            cls.build_send_payload("removeDOMStorageItem", {
                "storageId": storageId,
                "key": key,
            }),
            None
        )

    @classmethod
    def setDOMStorageItem(cls,
                          storageId: Union['StorageId'],
                          key: Union['str'],
                          value: Union['str'],
                          ):
        """
        :param storageId: 
        :type storageId: StorageId
        :param key: 
        :type key: str
        :param value: 
        :type value: str
        """
        return (
            cls.build_send_payload("setDOMStorageItem", {
                "storageId": storageId,
                "key": key,
                "value": value,
            }),
            None
        )



class DomStorageItemAddedEvent(BaseEvent):

    js_name = 'Domstorage.domStorageItemAdded'
    hashable = ['storageId']
    is_hashable = True

    def __init__(self,
                 storageId: Union['StorageId', dict],
                 key: Union['str', dict],
                 newValue: Union['str', dict],
                 ):
        if isinstance(storageId, dict):
            storageId = StorageId(**storageId)
        self.storageId = storageId
        if isinstance(key, dict):
            key = str(**key)
        self.key = key
        if isinstance(newValue, dict):
            newValue = str(**newValue)
        self.newValue = newValue

    @classmethod
    def build_hash(cls, storageId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DomStorageItemRemovedEvent(BaseEvent):

    js_name = 'Domstorage.domStorageItemRemoved'
    hashable = ['storageId']
    is_hashable = True

    def __init__(self,
                 storageId: Union['StorageId', dict],
                 key: Union['str', dict],
                 ):
        if isinstance(storageId, dict):
            storageId = StorageId(**storageId)
        self.storageId = storageId
        if isinstance(key, dict):
            key = str(**key)
        self.key = key

    @classmethod
    def build_hash(cls, storageId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DomStorageItemUpdatedEvent(BaseEvent):

    js_name = 'Domstorage.domStorageItemUpdated'
    hashable = ['storageId']
    is_hashable = True

    def __init__(self,
                 storageId: Union['StorageId', dict],
                 key: Union['str', dict],
                 oldValue: Union['str', dict],
                 newValue: Union['str', dict],
                 ):
        if isinstance(storageId, dict):
            storageId = StorageId(**storageId)
        self.storageId = storageId
        if isinstance(key, dict):
            key = str(**key)
        self.key = key
        if isinstance(oldValue, dict):
            oldValue = str(**oldValue)
        self.oldValue = oldValue
        if isinstance(newValue, dict):
            newValue = str(**newValue)
        self.newValue = newValue

    @classmethod
    def build_hash(cls, storageId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DomStorageItemsClearedEvent(BaseEvent):

    js_name = 'Domstorage.domStorageItemsCleared'
    hashable = ['storageId']
    is_hashable = True

    def __init__(self,
                 storageId: Union['StorageId', dict],
                 ):
        if isinstance(storageId, dict):
            storageId = StorageId(**storageId)
        self.storageId = storageId

    @classmethod
    def build_hash(cls, storageId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM

# StyleSheetId: 
StyleSheetId = str

# StyleSheetOrigin: Stylesheet type: "injected" for stylesheets injected via extension, "user-agent" for user-agentstylesheets, "inspector" for stylesheets created by the inspector (i.e. those holding the "viainspector" rules), "regular" for regular stylesheets.
StyleSheetOrigin = str

# PseudoElementMatches: CSS rule collection for a single pseudo style.
class PseudoElementMatches(ChromeTypeBase):
    def __init__(self,
                 pseudoType: Union['DOM.PseudoType'],
                 matches: Union['[RuleMatch]'],
                 ):

        self.pseudoType = pseudoType
        self.matches = matches


# InheritedStyleEntry: Inherited CSS rule collection from ancestor node.
class InheritedStyleEntry(ChromeTypeBase):
    def __init__(self,
                 matchedCSSRules: Union['[RuleMatch]'],
                 inlineStyle: Optional['CSSStyle'] = None,
                 ):

        self.inlineStyle = inlineStyle
        self.matchedCSSRules = matchedCSSRules


# RuleMatch: Match data for a CSS rule.
class RuleMatch(ChromeTypeBase):
    def __init__(self,
                 rule: Union['CSSRule'],
                 matchingSelectors: Union['[]'],
                 ):

        self.rule = rule
        self.matchingSelectors = matchingSelectors


# Value: Data for a simple selector (these are delimited by commas in a selector list).
class Value(ChromeTypeBase):
    def __init__(self,
                 text: Union['str'],
                 range: Optional['SourceRange'] = None,
                 ):

        self.text = text
        self.range = range


# SelectorList: Selector list data.
class SelectorList(ChromeTypeBase):
    def __init__(self,
                 selectors: Union['[Value]'],
                 text: Union['str'],
                 ):

        self.selectors = selectors
        self.text = text


# CSSStyleSheetHeader: CSS stylesheet metainformation.
class CSSStyleSheetHeader(ChromeTypeBase):
    def __init__(self,
                 styleSheetId: Union['StyleSheetId'],
                 frameId: Union['Page.FrameId'],
                 sourceURL: Union['str'],
                 origin: Union['StyleSheetOrigin'],
                 title: Union['str'],
                 disabled: Union['bool'],
                 isInline: Union['bool'],
                 startLine: Union['float'],
                 startColumn: Union['float'],
                 length: Union['float'],
                 sourceMapURL: Optional['str'] = None,
                 ownerNode: Optional['DOM.BackendNodeId'] = None,
                 hasSourceURL: Optional['bool'] = None,
                 ):

        self.styleSheetId = styleSheetId
        self.frameId = frameId
        self.sourceURL = sourceURL
        self.sourceMapURL = sourceMapURL
        self.origin = origin
        self.title = title
        self.ownerNode = ownerNode
        self.disabled = disabled
        self.hasSourceURL = hasSourceURL
        self.isInline = isInline
        self.startLine = startLine
        self.startColumn = startColumn
        self.length = length


# CSSRule: CSS rule representation.
class CSSRule(ChromeTypeBase):
    def __init__(self,
                 selectorList: Union['SelectorList'],
                 origin: Union['StyleSheetOrigin'],
                 style: Union['CSSStyle'],
                 styleSheetId: Optional['StyleSheetId'] = None,
                 media: Optional['[CSSMedia]'] = None,
                 ):

        self.styleSheetId = styleSheetId
        self.selectorList = selectorList
        self.origin = origin
        self.style = style
        self.media = media


# RuleUsage: CSS coverage information.
class RuleUsage(ChromeTypeBase):
    def __init__(self,
                 styleSheetId: Union['StyleSheetId'],
                 startOffset: Union['float'],
                 endOffset: Union['float'],
                 used: Union['bool'],
                 ):

        self.styleSheetId = styleSheetId
        self.startOffset = startOffset
        self.endOffset = endOffset
        self.used = used


# SourceRange: Text range within a resource. All numbers are zero-based.
class SourceRange(ChromeTypeBase):
    def __init__(self,
                 startLine: Union['int'],
                 startColumn: Union['int'],
                 endLine: Union['int'],
                 endColumn: Union['int'],
                 ):

        self.startLine = startLine
        self.startColumn = startColumn
        self.endLine = endLine
        self.endColumn = endColumn


# ShorthandEntry: 
class ShorthandEntry(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 important: Optional['bool'] = None,
                 ):

        self.name = name
        self.value = value
        self.important = important


# CSSComputedStyleProperty: 
class CSSComputedStyleProperty(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 ):

        self.name = name
        self.value = value


# CSSStyle: CSS style representation.
class CSSStyle(ChromeTypeBase):
    def __init__(self,
                 cssProperties: Union['[CSSProperty]'],
                 shorthandEntries: Union['[ShorthandEntry]'],
                 styleSheetId: Optional['StyleSheetId'] = None,
                 cssText: Optional['str'] = None,
                 range: Optional['SourceRange'] = None,
                 ):

        self.styleSheetId = styleSheetId
        self.cssProperties = cssProperties
        self.shorthandEntries = shorthandEntries
        self.cssText = cssText
        self.range = range


# CSSProperty: CSS property declaration data.
class CSSProperty(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 important: Optional['bool'] = None,
                 implicit: Optional['bool'] = None,
                 text: Optional['str'] = None,
                 parsedOk: Optional['bool'] = None,
                 disabled: Optional['bool'] = None,
                 range: Optional['SourceRange'] = None,
                 ):

        self.name = name
        self.value = value
        self.important = important
        self.implicit = implicit
        self.text = text
        self.parsedOk = parsedOk
        self.disabled = disabled
        self.range = range


# CSSMedia: CSS media rule descriptor.
class CSSMedia(ChromeTypeBase):
    def __init__(self,
                 text: Union['str'],
                 source: Union['str'],
                 sourceURL: Optional['str'] = None,
                 range: Optional['SourceRange'] = None,
                 styleSheetId: Optional['StyleSheetId'] = None,
                 mediaList: Optional['[MediaQuery]'] = None,
                 ):

        self.text = text
        self.source = source
        self.sourceURL = sourceURL
        self.range = range
        self.styleSheetId = styleSheetId
        self.mediaList = mediaList


# MediaQuery: Media query descriptor.
class MediaQuery(ChromeTypeBase):
    def __init__(self,
                 expressions: Union['[MediaQueryExpression]'],
                 active: Union['bool'],
                 ):

        self.expressions = expressions
        self.active = active


# MediaQueryExpression: Media query expression descriptor.
class MediaQueryExpression(ChromeTypeBase):
    def __init__(self,
                 value: Union['float'],
                 unit: Union['str'],
                 feature: Union['str'],
                 valueRange: Optional['SourceRange'] = None,
                 computedLength: Optional['float'] = None,
                 ):

        self.value = value
        self.unit = unit
        self.feature = feature
        self.valueRange = valueRange
        self.computedLength = computedLength


# PlatformFontUsage: Information about amount of glyphs that were rendered with given font.
class PlatformFontUsage(ChromeTypeBase):
    def __init__(self,
                 familyName: Union['str'],
                 isCustomFont: Union['bool'],
                 glyphCount: Union['float'],
                 ):

        self.familyName = familyName
        self.isCustomFont = isCustomFont
        self.glyphCount = glyphCount


# FontFace: Properties of a web font: https://www.w3.org/TR/2008/REC-CSS2-20080411/fonts.html#font-descriptions
class FontFace(ChromeTypeBase):
    def __init__(self,
                 fontFamily: Union['str'],
                 fontStyle: Union['str'],
                 fontVariant: Union['str'],
                 fontWeight: Union['str'],
                 fontStretch: Union['str'],
                 unicodeRange: Union['str'],
                 src: Union['str'],
                 platformFontFamily: Union['str'],
                 ):

        self.fontFamily = fontFamily
        self.fontStyle = fontStyle
        self.fontVariant = fontVariant
        self.fontWeight = fontWeight
        self.fontStretch = fontStretch
        self.unicodeRange = unicodeRange
        self.src = src
        self.platformFontFamily = platformFontFamily


# CSSKeyframesRule: CSS keyframes rule representation.
class CSSKeyframesRule(ChromeTypeBase):
    def __init__(self,
                 animationName: Union['Value'],
                 keyframes: Union['[CSSKeyframeRule]'],
                 ):

        self.animationName = animationName
        self.keyframes = keyframes


# CSSKeyframeRule: CSS keyframe rule representation.
class CSSKeyframeRule(ChromeTypeBase):
    def __init__(self,
                 origin: Union['StyleSheetOrigin'],
                 keyText: Union['Value'],
                 style: Union['CSSStyle'],
                 styleSheetId: Optional['StyleSheetId'] = None,
                 ):

        self.styleSheetId = styleSheetId
        self.origin = origin
        self.keyText = keyText
        self.style = style


# StyleDeclarationEdit: A descriptor of operation to mutate style declaration text.
class StyleDeclarationEdit(ChromeTypeBase):
    def __init__(self,
                 styleSheetId: Union['StyleSheetId'],
                 range: Union['SourceRange'],
                 text: Union['str'],
                 ):

        self.styleSheetId = styleSheetId
        self.range = range
        self.text = text


class CSS(PayloadMixin):
    """ This domain exposes CSS read/write operations. All CSS objects (stylesheets, rules, and styles)
have an associated `id` used in subsequent operations on the related object. Each object type has
a specific `id` structure, and those are not interchangeable between objects of different kinds.
CSS objects can be loaded using the `get*ForNode()` calls (which accept a DOM node id). A client
can also keep track of stylesheets via the `styleSheetAdded`/`styleSheetRemoved` events and
subsequently load the required stylesheet contents using the `getStyleSheet[Text]()` methods.
    """
    @classmethod
    def addRule(cls,
                styleSheetId: Union['StyleSheetId'],
                ruleText: Union['str'],
                location: Union['SourceRange'],
                ):
        """Inserts a new rule with the given `ruleText` in a stylesheet with given `styleSheetId`, at the
position specified by `location`.
        :param styleSheetId: The css style sheet identifier where a new rule should be inserted.
        :type styleSheetId: StyleSheetId
        :param ruleText: The text of a new rule.
        :type ruleText: str
        :param location: Text position of a new rule in the target style sheet.
        :type location: SourceRange
        """
        return (
            cls.build_send_payload("addRule", {
                "styleSheetId": styleSheetId,
                "ruleText": ruleText,
                "location": location,
            }),
            cls.convert_payload({
                "rule": {
                    "class": CSSRule,
                    "optional": False
                },
            })
        )

    @classmethod
    def collectClassNames(cls,
                          styleSheetId: Union['StyleSheetId'],
                          ):
        """Returns all class names from specified stylesheet.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        """
        return (
            cls.build_send_payload("collectClassNames", {
                "styleSheetId": styleSheetId,
            }),
            cls.convert_payload({
                "classNames": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def createStyleSheet(cls,
                         frameId: Union['Page.FrameId'],
                         ):
        """Creates a new special "via-inspector" stylesheet in the frame with given `frameId`.
        :param frameId: Identifier of the frame where "via-inspector" stylesheet should be created.
        :type frameId: Page.FrameId
        """
        return (
            cls.build_send_payload("createStyleSheet", {
                "frameId": frameId,
            }),
            cls.convert_payload({
                "styleSheetId": {
                    "class": StyleSheetId,
                    "optional": False
                },
            })
        )

    @classmethod
    def disable(cls):
        """Disables the CSS agent for the given page.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables the CSS agent for the given page. Clients should not assume that the CSS agent has been
enabled until the result of this command is received.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def forcePseudoState(cls,
                         nodeId: Union['DOM.NodeId'],
                         forcedPseudoClasses: Union['[]'],
                         ):
        """Ensures that the given node will have specified pseudo-classes whenever its style is computed by
the browser.
        :param nodeId: The element id for which to force the pseudo state.
        :type nodeId: DOM.NodeId
        :param forcedPseudoClasses: Element pseudo classes to force when computing the element's style.
        :type forcedPseudoClasses: []
        """
        return (
            cls.build_send_payload("forcePseudoState", {
                "nodeId": nodeId,
                "forcedPseudoClasses": forcedPseudoClasses,
            }),
            None
        )

    @classmethod
    def getBackgroundColors(cls,
                            nodeId: Union['DOM.NodeId'],
                            ):
        """
        :param nodeId: Id of the node to get background colors for.
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getBackgroundColors", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "backgroundColors": {
                    "class": [],
                    "optional": True
                },
                "computedFontSize": {
                    "class": str,
                    "optional": True
                },
                "computedFontWeight": {
                    "class": str,
                    "optional": True
                },
            })
        )

    @classmethod
    def getComputedStyleForNode(cls,
                                nodeId: Union['DOM.NodeId'],
                                ):
        """Returns the computed style for a DOM node identified by `nodeId`.
        :param nodeId: 
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getComputedStyleForNode", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "computedStyle": {
                    "class": [CSSComputedStyleProperty],
                    "optional": False
                },
            })
        )

    @classmethod
    def getInlineStylesForNode(cls,
                               nodeId: Union['DOM.NodeId'],
                               ):
        """Returns the styles defined inline (explicitly in the "style" attribute and implicitly, using DOM
attributes) for a DOM node identified by `nodeId`.
        :param nodeId: 
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getInlineStylesForNode", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "inlineStyle": {
                    "class": CSSStyle,
                    "optional": True
                },
                "attributesStyle": {
                    "class": CSSStyle,
                    "optional": True
                },
            })
        )

    @classmethod
    def getMatchedStylesForNode(cls,
                                nodeId: Union['DOM.NodeId'],
                                ):
        """Returns requested styles for a DOM node identified by `nodeId`.
        :param nodeId: 
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getMatchedStylesForNode", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "inlineStyle": {
                    "class": CSSStyle,
                    "optional": True
                },
                "attributesStyle": {
                    "class": CSSStyle,
                    "optional": True
                },
                "matchedCSSRules": {
                    "class": [RuleMatch],
                    "optional": True
                },
                "pseudoElements": {
                    "class": [PseudoElementMatches],
                    "optional": True
                },
                "inherited": {
                    "class": [InheritedStyleEntry],
                    "optional": True
                },
                "cssKeyframesRules": {
                    "class": [CSSKeyframesRule],
                    "optional": True
                },
            })
        )

    @classmethod
    def getMediaQueries(cls):
        """Returns all media queries parsed by the rendering engine.
        """
        return (
            cls.build_send_payload("getMediaQueries", {
            }),
            cls.convert_payload({
                "medias": {
                    "class": [CSSMedia],
                    "optional": False
                },
            })
        )

    @classmethod
    def getPlatformFontsForNode(cls,
                                nodeId: Union['DOM.NodeId'],
                                ):
        """Requests information about platform fonts which we used to render child TextNodes in the given
node.
        :param nodeId: 
        :type nodeId: DOM.NodeId
        """
        return (
            cls.build_send_payload("getPlatformFontsForNode", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "fonts": {
                    "class": [PlatformFontUsage],
                    "optional": False
                },
            })
        )

    @classmethod
    def getStyleSheetText(cls,
                          styleSheetId: Union['StyleSheetId'],
                          ):
        """Returns the current textual content for a stylesheet.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        """
        return (
            cls.build_send_payload("getStyleSheetText", {
                "styleSheetId": styleSheetId,
            }),
            cls.convert_payload({
                "text": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def setEffectivePropertyValueForNode(cls,
                                         nodeId: Union['DOM.NodeId'],
                                         propertyName: Union['str'],
                                         value: Union['str'],
                                         ):
        """Find a rule with the given active property for the given node and set the new value for this
property
        :param nodeId: The element id for which to set property.
        :type nodeId: DOM.NodeId
        :param propertyName: 
        :type propertyName: str
        :param value: 
        :type value: str
        """
        return (
            cls.build_send_payload("setEffectivePropertyValueForNode", {
                "nodeId": nodeId,
                "propertyName": propertyName,
                "value": value,
            }),
            None
        )

    @classmethod
    def setKeyframeKey(cls,
                       styleSheetId: Union['StyleSheetId'],
                       range: Union['SourceRange'],
                       keyText: Union['str'],
                       ):
        """Modifies the keyframe rule key text.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        :param range: 
        :type range: SourceRange
        :param keyText: 
        :type keyText: str
        """
        return (
            cls.build_send_payload("setKeyframeKey", {
                "styleSheetId": styleSheetId,
                "range": range,
                "keyText": keyText,
            }),
            cls.convert_payload({
                "keyText": {
                    "class": Value,
                    "optional": False
                },
            })
        )

    @classmethod
    def setMediaText(cls,
                     styleSheetId: Union['StyleSheetId'],
                     range: Union['SourceRange'],
                     text: Union['str'],
                     ):
        """Modifies the rule selector.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        :param range: 
        :type range: SourceRange
        :param text: 
        :type text: str
        """
        return (
            cls.build_send_payload("setMediaText", {
                "styleSheetId": styleSheetId,
                "range": range,
                "text": text,
            }),
            cls.convert_payload({
                "media": {
                    "class": CSSMedia,
                    "optional": False
                },
            })
        )

    @classmethod
    def setRuleSelector(cls,
                        styleSheetId: Union['StyleSheetId'],
                        range: Union['SourceRange'],
                        selector: Union['str'],
                        ):
        """Modifies the rule selector.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        :param range: 
        :type range: SourceRange
        :param selector: 
        :type selector: str
        """
        return (
            cls.build_send_payload("setRuleSelector", {
                "styleSheetId": styleSheetId,
                "range": range,
                "selector": selector,
            }),
            cls.convert_payload({
                "selectorList": {
                    "class": SelectorList,
                    "optional": False
                },
            })
        )

    @classmethod
    def setStyleSheetText(cls,
                          styleSheetId: Union['StyleSheetId'],
                          text: Union['str'],
                          ):
        """Sets the new stylesheet text.
        :param styleSheetId: 
        :type styleSheetId: StyleSheetId
        :param text: 
        :type text: str
        """
        return (
            cls.build_send_payload("setStyleSheetText", {
                "styleSheetId": styleSheetId,
                "text": text,
            }),
            cls.convert_payload({
                "sourceMapURL": {
                    "class": str,
                    "optional": True
                },
            })
        )

    @classmethod
    def setStyleTexts(cls,
                      edits: Union['[StyleDeclarationEdit]'],
                      ):
        """Applies specified style edits one after another in the given order.
        :param edits: 
        :type edits: [StyleDeclarationEdit]
        """
        return (
            cls.build_send_payload("setStyleTexts", {
                "edits": edits,
            }),
            cls.convert_payload({
                "styles": {
                    "class": [CSSStyle],
                    "optional": False
                },
            })
        )

    @classmethod
    def startRuleUsageTracking(cls):
        """Enables the selector recording.
        """
        return (
            cls.build_send_payload("startRuleUsageTracking", {
            }),
            None
        )

    @classmethod
    def stopRuleUsageTracking(cls):
        """Stop tracking rule usage and return the list of rules that were used since last call to
`takeCoverageDelta` (or since start of coverage instrumentation)
        """
        return (
            cls.build_send_payload("stopRuleUsageTracking", {
            }),
            cls.convert_payload({
                "ruleUsage": {
                    "class": [RuleUsage],
                    "optional": False
                },
            })
        )

    @classmethod
    def takeCoverageDelta(cls):
        """Obtain list of rules that became used since last call to this method (or since start of coverage
instrumentation)
        """
        return (
            cls.build_send_payload("takeCoverageDelta", {
            }),
            cls.convert_payload({
                "coverage": {
                    "class": [RuleUsage],
                    "optional": False
                },
            })
        )



class FontsUpdatedEvent(BaseEvent):

    js_name = 'Css.fontsUpdated'
    hashable = []
    is_hashable = False

    def __init__(self,
                 font: Union['FontFace', dict, None] = None,
                 ):
        if isinstance(font, dict):
            font = FontFace(**font)
        self.font = font

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class MediaQueryResultChangedEvent(BaseEvent):

    js_name = 'Css.mediaQueryResultChanged'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class StyleSheetAddedEvent(BaseEvent):

    js_name = 'Css.styleSheetAdded'
    hashable = []
    is_hashable = False

    def __init__(self,
                 header: Union['CSSStyleSheetHeader', dict],
                 ):
        if isinstance(header, dict):
            header = CSSStyleSheetHeader(**header)
        self.header = header

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class StyleSheetChangedEvent(BaseEvent):

    js_name = 'Css.styleSheetChanged'
    hashable = ['styleSheetId']
    is_hashable = True

    def __init__(self,
                 styleSheetId: Union['StyleSheetId', dict],
                 ):
        if isinstance(styleSheetId, dict):
            styleSheetId = StyleSheetId(**styleSheetId)
        self.styleSheetId = styleSheetId

    @classmethod
    def build_hash(cls, styleSheetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class StyleSheetRemovedEvent(BaseEvent):

    js_name = 'Css.styleSheetRemoved'
    hashable = ['styleSheetId']
    is_hashable = True

    def __init__(self,
                 styleSheetId: Union['StyleSheetId', dict],
                 ):
        if isinstance(styleSheetId, dict):
            styleSheetId = StyleSheetId(**styleSheetId)
        self.styleSheetId = styleSheetId

    @classmethod
    def build_hash(cls, styleSheetId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import page as Page

class Testing(PayloadMixin):
    """ Testing domain is a dumping ground for the capabilities requires for browser or app testing that do not fit other
domains.
    """
    @classmethod
    def generateTestReport(cls,
                           message: Union['str'],
                           group: Optional['str'] = None,
                           ):
        """Generates a report for testing.
        :param message: Message to be displayed in the report.
        :type message: str
        :param group: Specifies the endpoint group to deliver the report to.
        :type group: str
        """
        return (
            cls.build_send_payload("generateTestReport", {
                "message": message,
                "group": group,
            }),
            None
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime
from chromewhip.protocol import dom as DOM

# Animation: Animation instance.
class Animation(ChromeTypeBase):
    def __init__(self,
                 id: Union['str'],
                 name: Union['str'],
                 pausedState: Union['bool'],
                 playState: Union['str'],
                 playbackRate: Union['float'],
                 startTime: Union['float'],
                 currentTime: Union['float'],
                 type: Union['str'],
                 source: Optional['AnimationEffect'] = None,
                 cssId: Optional['str'] = None,
                 ):

        self.id = id
        self.name = name
        self.pausedState = pausedState
        self.playState = playState
        self.playbackRate = playbackRate
        self.startTime = startTime
        self.currentTime = currentTime
        self.type = type
        self.source = source
        self.cssId = cssId


# AnimationEffect: AnimationEffect instance
class AnimationEffect(ChromeTypeBase):
    def __init__(self,
                 delay: Union['float'],
                 endDelay: Union['float'],
                 iterationStart: Union['float'],
                 iterations: Union['float'],
                 duration: Union['float'],
                 direction: Union['str'],
                 fill: Union['str'],
                 easing: Union['str'],
                 backendNodeId: Optional['DOM.BackendNodeId'] = None,
                 keyframesRule: Optional['KeyframesRule'] = None,
                 ):

        self.delay = delay
        self.endDelay = endDelay
        self.iterationStart = iterationStart
        self.iterations = iterations
        self.duration = duration
        self.direction = direction
        self.fill = fill
        self.backendNodeId = backendNodeId
        self.keyframesRule = keyframesRule
        self.easing = easing


# KeyframesRule: Keyframes Rule
class KeyframesRule(ChromeTypeBase):
    def __init__(self,
                 keyframes: Union['[KeyframeStyle]'],
                 name: Optional['str'] = None,
                 ):

        self.name = name
        self.keyframes = keyframes


# KeyframeStyle: Keyframe Style
class KeyframeStyle(ChromeTypeBase):
    def __init__(self,
                 offset: Union['str'],
                 easing: Union['str'],
                 ):

        self.offset = offset
        self.easing = easing


class Animation(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disables animation domain notifications.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables animation domain notifications.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getCurrentTime(cls,
                       id: Union['str'],
                       ):
        """Returns the current time of the an animation.
        :param id: Id of animation.
        :type id: str
        """
        return (
            cls.build_send_payload("getCurrentTime", {
                "id": id,
            }),
            cls.convert_payload({
                "currentTime": {
                    "class": float,
                    "optional": False
                },
            })
        )

    @classmethod
    def getPlaybackRate(cls):
        """Gets the playback rate of the document timeline.
        """
        return (
            cls.build_send_payload("getPlaybackRate", {
            }),
            cls.convert_payload({
                "playbackRate": {
                    "class": float,
                    "optional": False
                },
            })
        )

    @classmethod
    def releaseAnimations(cls,
                          animations: Union['[]'],
                          ):
        """Releases a set of animations to no longer be manipulated.
        :param animations: List of animation ids to seek.
        :type animations: []
        """
        return (
            cls.build_send_payload("releaseAnimations", {
                "animations": animations,
            }),
            None
        )

    @classmethod
    def resolveAnimation(cls,
                         animationId: Union['str'],
                         ):
        """Gets the remote object of the Animation.
        :param animationId: Animation id.
        :type animationId: str
        """
        return (
            cls.build_send_payload("resolveAnimation", {
                "animationId": animationId,
            }),
            cls.convert_payload({
                "remoteObject": {
                    "class": Runtime.RemoteObject,
                    "optional": False
                },
            })
        )

    @classmethod
    def seekAnimations(cls,
                       animations: Union['[]'],
                       currentTime: Union['float'],
                       ):
        """Seek a set of animations to a particular time within each animation.
        :param animations: List of animation ids to seek.
        :type animations: []
        :param currentTime: Set the current time of each animation.
        :type currentTime: float
        """
        return (
            cls.build_send_payload("seekAnimations", {
                "animations": animations,
                "currentTime": currentTime,
            }),
            None
        )

    @classmethod
    def setPaused(cls,
                  animations: Union['[]'],
                  paused: Union['bool'],
                  ):
        """Sets the paused state of a set of animations.
        :param animations: Animations to set the pause state of.
        :type animations: []
        :param paused: Paused state to set to.
        :type paused: bool
        """
        return (
            cls.build_send_payload("setPaused", {
                "animations": animations,
                "paused": paused,
            }),
            None
        )

    @classmethod
    def setPlaybackRate(cls,
                        playbackRate: Union['float'],
                        ):
        """Sets the playback rate of the document timeline.
        :param playbackRate: Playback rate for animations on page
        :type playbackRate: float
        """
        return (
            cls.build_send_payload("setPlaybackRate", {
                "playbackRate": playbackRate,
            }),
            None
        )

    @classmethod
    def setTiming(cls,
                  animationId: Union['str'],
                  duration: Union['float'],
                  delay: Union['float'],
                  ):
        """Sets the timing of an animation node.
        :param animationId: Animation id.
        :type animationId: str
        :param duration: Duration of the animation.
        :type duration: float
        :param delay: Delay of the animation.
        :type delay: float
        """
        return (
            cls.build_send_payload("setTiming", {
                "animationId": animationId,
                "duration": duration,
                "delay": delay,
            }),
            None
        )



class AnimationCanceledEvent(BaseEvent):

    js_name = 'Animation.animationCanceled'
    hashable = ['id']
    is_hashable = True

    def __init__(self,
                 id: Union['str', dict],
                 ):
        if isinstance(id, dict):
            id = str(**id)
        self.id = id

    @classmethod
    def build_hash(cls, id):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class AnimationCreatedEvent(BaseEvent):

    js_name = 'Animation.animationCreated'
    hashable = ['id']
    is_hashable = True

    def __init__(self,
                 id: Union['str', dict],
                 ):
        if isinstance(id, dict):
            id = str(**id)
        self.id = id

    @classmethod
    def build_hash(cls, id):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class AnimationStartedEvent(BaseEvent):

    js_name = 'Animation.animationStarted'
    hashable = ['animationId']
    is_hashable = True

    def __init__(self,
                 animation: Union['Animation', dict],
                 ):
        if isinstance(animation, dict):
            animation = Animation(**animation)
        self.animation = animation

    @classmethod
    def build_hash(cls, animationId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime

# NodeId: Unique DOM node identifier.
NodeId = int

# BackendNodeId: Unique DOM node identifier used to reference a node that may not have been pushed to thefront-end.
BackendNodeId = int

# BackendNode: Backend node with a friendly name.
class BackendNode(ChromeTypeBase):
    def __init__(self,
                 nodeType: Union['int'],
                 nodeName: Union['str'],
                 backendNodeId: Union['BackendNodeId'],
                 ):

        self.nodeType = nodeType
        self.nodeName = nodeName
        self.backendNodeId = backendNodeId


# PseudoType: Pseudo element type.
PseudoType = str

# ShadowRootType: Shadow root type.
ShadowRootType = str

# Node: DOM interaction is implemented in terms of mirror objects that represent the actual DOM nodes.DOMNode is a base node mirror type.
class Node(ChromeTypeBase):
    def __init__(self,
                 nodeId: Union['NodeId'],
                 backendNodeId: Union['BackendNodeId'],
                 nodeType: Union['int'],
                 nodeName: Union['str'],
                 localName: Union['str'],
                 nodeValue: Union['str'],
                 parentId: Optional['NodeId'] = None,
                 childNodeCount: Optional['int'] = None,
                 children: Optional['[Node]'] = None,
                 attributes: Optional['[]'] = None,
                 documentURL: Optional['str'] = None,
                 baseURL: Optional['str'] = None,
                 publicId: Optional['str'] = None,
                 systemId: Optional['str'] = None,
                 internalSubset: Optional['str'] = None,
                 xmlVersion: Optional['str'] = None,
                 name: Optional['str'] = None,
                 value: Optional['str'] = None,
                 pseudoType: Optional['PseudoType'] = None,
                 shadowRootType: Optional['ShadowRootType'] = None,
                 frameId: Optional['Page.FrameId'] = None,
                 contentDocument: Optional['Node'] = None,
                 shadowRoots: Optional['[Node]'] = None,
                 templateContent: Optional['Node'] = None,
                 pseudoElements: Optional['[Node]'] = None,
                 importedDocument: Optional['Node'] = None,
                 distributedNodes: Optional['[BackendNode]'] = None,
                 isSVG: Optional['bool'] = None,
                 ):

        self.nodeId = nodeId
        self.parentId = parentId
        self.backendNodeId = backendNodeId
        self.nodeType = nodeType
        self.nodeName = nodeName
        self.localName = localName
        self.nodeValue = nodeValue
        self.childNodeCount = childNodeCount
        self.children = children
        self.attributes = attributes
        self.documentURL = documentURL
        self.baseURL = baseURL
        self.publicId = publicId
        self.systemId = systemId
        self.internalSubset = internalSubset
        self.xmlVersion = xmlVersion
        self.name = name
        self.value = value
        self.pseudoType = pseudoType
        self.shadowRootType = shadowRootType
        self.frameId = frameId
        self.contentDocument = contentDocument
        self.shadowRoots = shadowRoots
        self.templateContent = templateContent
        self.pseudoElements = pseudoElements
        self.importedDocument = importedDocument
        self.distributedNodes = distributedNodes
        self.isSVG = isSVG


# RGBA: A structure holding an RGBA color.
class RGBA(ChromeTypeBase):
    def __init__(self,
                 r: Union['int'],
                 g: Union['int'],
                 b: Union['int'],
                 a: Optional['float'] = None,
                 ):

        self.r = r
        self.g = g
        self.b = b
        self.a = a


# Quad: An array of quad vertices, x immediately followed by y for each point, points clock-wise.
Quad = [float]

# BoxModel: Box model.
class BoxModel(ChromeTypeBase):
    def __init__(self,
                 content: Union['Quad'],
                 padding: Union['Quad'],
                 border: Union['Quad'],
                 margin: Union['Quad'],
                 width: Union['int'],
                 height: Union['int'],
                 shapeOutside: Optional['ShapeOutsideInfo'] = None,
                 ):

        self.content = content
        self.padding = padding
        self.border = border
        self.margin = margin
        self.width = width
        self.height = height
        self.shapeOutside = shapeOutside


# ShapeOutsideInfo: CSS Shape Outside details.
class ShapeOutsideInfo(ChromeTypeBase):
    def __init__(self,
                 bounds: Union['Quad'],
                 shape: Union['[]'],
                 marginShape: Union['[]'],
                 ):

        self.bounds = bounds
        self.shape = shape
        self.marginShape = marginShape


# Rect: Rectangle.
class Rect(ChromeTypeBase):
    def __init__(self,
                 x: Union['float'],
                 y: Union['float'],
                 width: Union['float'],
                 height: Union['float'],
                 ):

        self.x = x
        self.y = y
        self.width = width
        self.height = height


class DOM(PayloadMixin):
    """ This domain exposes DOM read/write operations. Each DOM Node is represented with its mirror object
that has an `id`. This `id` can be used to get additional information on the Node, resolve it into
the JavaScript object wrapper, etc. It is important that client receives DOM events only for the
nodes that are known to the client. Backend keeps track of the nodes that were sent to the client
and never sends the same node twice. It is client's responsibility to collect information about
the nodes that were sent to the client.<p>Note that `iframe` owner elements will return
corresponding document elements as their child nodes.</p>
    """
    @classmethod
    def collectClassNamesFromSubtree(cls,
                                     nodeId: Union['NodeId'],
                                     ):
        """Collects class names for the node with given id and all of it's child nodes.
        :param nodeId: Id of the node to collect class names.
        :type nodeId: NodeId
        """
        return (
            cls.build_send_payload("collectClassNamesFromSubtree", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "classNames": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def copyTo(cls,
               nodeId: Union['NodeId'],
               targetNodeId: Union['NodeId'],
               insertBeforeNodeId: Optional['NodeId'] = None,
               ):
        """Creates a deep copy of the specified node and places it into the target container before the
given anchor.
        :param nodeId: Id of the node to copy.
        :type nodeId: NodeId
        :param targetNodeId: Id of the element to drop the copy into.
        :type targetNodeId: NodeId
        :param insertBeforeNodeId: Drop the copy before this node (if absent, the copy becomes the last child of
`targetNodeId`).
        :type insertBeforeNodeId: NodeId
        """
        return (
            cls.build_send_payload("copyTo", {
                "nodeId": nodeId,
                "targetNodeId": targetNodeId,
                "insertBeforeNodeId": insertBeforeNodeId,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def describeNode(cls,
                     nodeId: Optional['NodeId'] = None,
                     backendNodeId: Optional['BackendNodeId'] = None,
                     objectId: Optional['Runtime.RemoteObjectId'] = None,
                     depth: Optional['int'] = None,
                     pierce: Optional['bool'] = None,
                     ):
        """Describes node given its id, does not require domain to be enabled. Does not start tracking any
objects, can be used for automation.
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        :param depth: The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the
entire subtree or provide an integer larger than 0.
        :type depth: int
        :param pierce: Whether or not iframes and shadow roots should be traversed when returning the subtree
(default is false).
        :type pierce: bool
        """
        return (
            cls.build_send_payload("describeNode", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
                "depth": depth,
                "pierce": pierce,
            }),
            cls.convert_payload({
                "node": {
                    "class": Node,
                    "optional": False
                },
            })
        )

    @classmethod
    def disable(cls):
        """Disables DOM agent for the given page.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def discardSearchResults(cls,
                             searchId: Union['str'],
                             ):
        """Discards search results from the session with the given id. `getSearchResults` should no longer
be called for that search.
        :param searchId: Unique search session identifier.
        :type searchId: str
        """
        return (
            cls.build_send_payload("discardSearchResults", {
                "searchId": searchId,
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables DOM agent for the given page.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def focus(cls,
              nodeId: Optional['NodeId'] = None,
              backendNodeId: Optional['BackendNodeId'] = None,
              objectId: Optional['Runtime.RemoteObjectId'] = None,
              ):
        """Focuses the given element.
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("focus", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            None
        )

    @classmethod
    def getAttributes(cls,
                      nodeId: Union['NodeId'],
                      ):
        """Returns attributes for the specified node.
        :param nodeId: Id of the node to retrieve attibutes for.
        :type nodeId: NodeId
        """
        return (
            cls.build_send_payload("getAttributes", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "attributes": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def getBoxModel(cls,
                    nodeId: Optional['NodeId'] = None,
                    backendNodeId: Optional['BackendNodeId'] = None,
                    objectId: Optional['Runtime.RemoteObjectId'] = None,
                    ):
        """Returns boxes for the given node.
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("getBoxModel", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            cls.convert_payload({
                "model": {
                    "class": BoxModel,
                    "optional": False
                },
            })
        )

    @classmethod
    def getContentQuads(cls,
                        nodeId: Optional['NodeId'] = None,
                        backendNodeId: Optional['BackendNodeId'] = None,
                        objectId: Optional['Runtime.RemoteObjectId'] = None,
                        ):
        """Returns quads that describe node position on the page. This method
might return multiple quads for inline nodes.
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("getContentQuads", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            cls.convert_payload({
                "quads": {
                    "class": [Quad],
                    "optional": False
                },
            })
        )

    @classmethod
    def getDocument(cls,
                    depth: Optional['int'] = None,
                    pierce: Optional['bool'] = None,
                    ):
        """Returns the root DOM node (and optionally the subtree) to the caller.
        :param depth: The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the
entire subtree or provide an integer larger than 0.
        :type depth: int
        :param pierce: Whether or not iframes and shadow roots should be traversed when returning the subtree
(default is false).
        :type pierce: bool
        """
        return (
            cls.build_send_payload("getDocument", {
                "depth": depth,
                "pierce": pierce,
            }),
            cls.convert_payload({
                "root": {
                    "class": Node,
                    "optional": False
                },
            })
        )

    @classmethod
    def getFlattenedDocument(cls,
                             depth: Optional['int'] = None,
                             pierce: Optional['bool'] = None,
                             ):
        """Returns the root DOM node (and optionally the subtree) to the caller.
        :param depth: The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the
entire subtree or provide an integer larger than 0.
        :type depth: int
        :param pierce: Whether or not iframes and shadow roots should be traversed when returning the subtree
(default is false).
        :type pierce: bool
        """
        return (
            cls.build_send_payload("getFlattenedDocument", {
                "depth": depth,
                "pierce": pierce,
            }),
            cls.convert_payload({
                "nodes": {
                    "class": [Node],
                    "optional": False
                },
            })
        )

    @classmethod
    def getNodeForLocation(cls,
                           x: Union['int'],
                           y: Union['int'],
                           includeUserAgentShadowDOM: Optional['bool'] = None,
                           ):
        """Returns node id at given location. Depending on whether DOM domain is enabled, nodeId is
either returned or not.
        :param x: X coordinate.
        :type x: int
        :param y: Y coordinate.
        :type y: int
        :param includeUserAgentShadowDOM: False to skip to the nearest non-UA shadow root ancestor (default: false).
        :type includeUserAgentShadowDOM: bool
        """
        return (
            cls.build_send_payload("getNodeForLocation", {
                "x": x,
                "y": y,
                "includeUserAgentShadowDOM": includeUserAgentShadowDOM,
            }),
            cls.convert_payload({
                "backendNodeId": {
                    "class": BackendNodeId,
                    "optional": False
                },
                "nodeId": {
                    "class": NodeId,
                    "optional": True
                },
            })
        )

    @classmethod
    def getOuterHTML(cls,
                     nodeId: Optional['NodeId'] = None,
                     backendNodeId: Optional['BackendNodeId'] = None,
                     objectId: Optional['Runtime.RemoteObjectId'] = None,
                     ):
        """Returns node's HTML markup.
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("getOuterHTML", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            cls.convert_payload({
                "outerHTML": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def getRelayoutBoundary(cls,
                            nodeId: Union['NodeId'],
                            ):
        """Returns the id of the nearest ancestor that is a relayout boundary.
        :param nodeId: Id of the node.
        :type nodeId: NodeId
        """
        return (
            cls.build_send_payload("getRelayoutBoundary", {
                "nodeId": nodeId,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def getSearchResults(cls,
                         searchId: Union['str'],
                         fromIndex: Union['int'],
                         toIndex: Union['int'],
                         ):
        """Returns search results from given `fromIndex` to given `toIndex` from the search with the given
identifier.
        :param searchId: Unique search session identifier.
        :type searchId: str
        :param fromIndex: Start index of the search result to be returned.
        :type fromIndex: int
        :param toIndex: End index of the search result to be returned.
        :type toIndex: int
        """
        return (
            cls.build_send_payload("getSearchResults", {
                "searchId": searchId,
                "fromIndex": fromIndex,
                "toIndex": toIndex,
            }),
            cls.convert_payload({
                "nodeIds": {
                    "class": [NodeId],
                    "optional": False
                },
            })
        )

    @classmethod
    def hideHighlight(cls):
        """Hides any highlight.
        """
        return (
            cls.build_send_payload("hideHighlight", {
            }),
            None
        )

    @classmethod
    def highlightNode(cls):
        """Highlights DOM node.
        """
        return (
            cls.build_send_payload("highlightNode", {
            }),
            None
        )

    @classmethod
    def highlightRect(cls):
        """Highlights given rectangle.
        """
        return (
            cls.build_send_payload("highlightRect", {
            }),
            None
        )

    @classmethod
    def markUndoableState(cls):
        """Marks last undoable state.
        """
        return (
            cls.build_send_payload("markUndoableState", {
            }),
            None
        )

    @classmethod
    def moveTo(cls,
               nodeId: Union['NodeId'],
               targetNodeId: Union['NodeId'],
               insertBeforeNodeId: Optional['NodeId'] = None,
               ):
        """Moves node into the new container, places it before the given anchor.
        :param nodeId: Id of the node to move.
        :type nodeId: NodeId
        :param targetNodeId: Id of the element to drop the moved node into.
        :type targetNodeId: NodeId
        :param insertBeforeNodeId: Drop node before this one (if absent, the moved node becomes the last child of
`targetNodeId`).
        :type insertBeforeNodeId: NodeId
        """
        return (
            cls.build_send_payload("moveTo", {
                "nodeId": nodeId,
                "targetNodeId": targetNodeId,
                "insertBeforeNodeId": insertBeforeNodeId,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def performSearch(cls,
                      query: Union['str'],
                      includeUserAgentShadowDOM: Optional['bool'] = None,
                      ):
        """Searches for a given string in the DOM tree. Use `getSearchResults` to access search results or
`cancelSearch` to end this search session.
        :param query: Plain text or query selector or XPath search query.
        :type query: str
        :param includeUserAgentShadowDOM: True to search in user agent shadow DOM.
        :type includeUserAgentShadowDOM: bool
        """
        return (
            cls.build_send_payload("performSearch", {
                "query": query,
                "includeUserAgentShadowDOM": includeUserAgentShadowDOM,
            }),
            cls.convert_payload({
                "searchId": {
                    "class": str,
                    "optional": False
                },
                "resultCount": {
                    "class": int,
                    "optional": False
                },
            })
        )

    @classmethod
    def pushNodeByPathToFrontend(cls,
                                 path: Union['str'],
                                 ):
        """Requests that the node is sent to the caller given its path. // FIXME, use XPath
        :param path: Path to node in the proprietary format.
        :type path: str
        """
        return (
            cls.build_send_payload("pushNodeByPathToFrontend", {
                "path": path,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def pushNodesByBackendIdsToFrontend(cls,
                                        backendNodeIds: Union['[BackendNodeId]'],
                                        ):
        """Requests that a batch of nodes is sent to the caller given their backend node ids.
        :param backendNodeIds: The array of backend node ids.
        :type backendNodeIds: [BackendNodeId]
        """
        return (
            cls.build_send_payload("pushNodesByBackendIdsToFrontend", {
                "backendNodeIds": backendNodeIds,
            }),
            cls.convert_payload({
                "nodeIds": {
                    "class": [NodeId],
                    "optional": False
                },
            })
        )

    @classmethod
    def querySelector(cls,
                      nodeId: Union['NodeId'],
                      selector: Union['str'],
                      ):
        """Executes `querySelector` on a given node.
        :param nodeId: Id of the node to query upon.
        :type nodeId: NodeId
        :param selector: Selector string.
        :type selector: str
        """
        return (
            cls.build_send_payload("querySelector", {
                "nodeId": nodeId,
                "selector": selector,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def querySelectorAll(cls,
                         nodeId: Union['NodeId'],
                         selector: Union['str'],
                         ):
        """Executes `querySelectorAll` on a given node.
        :param nodeId: Id of the node to query upon.
        :type nodeId: NodeId
        :param selector: Selector string.
        :type selector: str
        """
        return (
            cls.build_send_payload("querySelectorAll", {
                "nodeId": nodeId,
                "selector": selector,
            }),
            cls.convert_payload({
                "nodeIds": {
                    "class": [NodeId],
                    "optional": False
                },
            })
        )

    @classmethod
    def redo(cls):
        """Re-does the last undone action.
        """
        return (
            cls.build_send_payload("redo", {
            }),
            None
        )

    @classmethod
    def removeAttribute(cls,
                        nodeId: Union['NodeId'],
                        name: Union['str'],
                        ):
        """Removes attribute with given name from an element with given id.
        :param nodeId: Id of the element to remove attribute from.
        :type nodeId: NodeId
        :param name: Name of the attribute to remove.
        :type name: str
        """
        return (
            cls.build_send_payload("removeAttribute", {
                "nodeId": nodeId,
                "name": name,
            }),
            None
        )

    @classmethod
    def removeNode(cls,
                   nodeId: Union['NodeId'],
                   ):
        """Removes node with given id.
        :param nodeId: Id of the node to remove.
        :type nodeId: NodeId
        """
        return (
            cls.build_send_payload("removeNode", {
                "nodeId": nodeId,
            }),
            None
        )

    @classmethod
    def requestChildNodes(cls,
                          nodeId: Union['NodeId'],
                          depth: Optional['int'] = None,
                          pierce: Optional['bool'] = None,
                          ):
        """Requests that children of the node with given id are returned to the caller in form of
`setChildNodes` events where not only immediate children are retrieved, but all children down to
the specified depth.
        :param nodeId: Id of the node to get children for.
        :type nodeId: NodeId
        :param depth: The maximum depth at which children should be retrieved, defaults to 1. Use -1 for the
entire subtree or provide an integer larger than 0.
        :type depth: int
        :param pierce: Whether or not iframes and shadow roots should be traversed when returning the sub-tree
(default is false).
        :type pierce: bool
        """
        return (
            cls.build_send_payload("requestChildNodes", {
                "nodeId": nodeId,
                "depth": depth,
                "pierce": pierce,
            }),
            None
        )

    @classmethod
    def requestNode(cls,
                    objectId: Union['Runtime.RemoteObjectId'],
                    ):
        """Requests that the node is sent to the caller given the JavaScript node object reference. All
nodes that form the path from the node to the root are also sent to the client as a series of
`setChildNodes` notifications.
        :param objectId: JavaScript object id to convert into node.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("requestNode", {
                "objectId": objectId,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def resolveNode(cls,
                    nodeId: Optional['NodeId'] = None,
                    backendNodeId: Optional['DOM.BackendNodeId'] = None,
                    objectGroup: Optional['str'] = None,
                    executionContextId: Optional['Runtime.ExecutionContextId'] = None,
                    ):
        """Resolves the JavaScript node object for a given NodeId or BackendNodeId.
        :param nodeId: Id of the node to resolve.
        :type nodeId: NodeId
        :param backendNodeId: Backend identifier of the node to resolve.
        :type backendNodeId: DOM.BackendNodeId
        :param objectGroup: Symbolic group name that can be used to release multiple objects.
        :type objectGroup: str
        :param executionContextId: Execution context in which to resolve the node.
        :type executionContextId: Runtime.ExecutionContextId
        """
        return (
            cls.build_send_payload("resolveNode", {
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectGroup": objectGroup,
                "executionContextId": executionContextId,
            }),
            cls.convert_payload({
                "object": {
                    "class": Runtime.RemoteObject,
                    "optional": False
                },
            })
        )

    @classmethod
    def setAttributeValue(cls,
                          nodeId: Union['NodeId'],
                          name: Union['str'],
                          value: Union['str'],
                          ):
        """Sets attribute for an element with given id.
        :param nodeId: Id of the element to set attribute for.
        :type nodeId: NodeId
        :param name: Attribute name.
        :type name: str
        :param value: Attribute value.
        :type value: str
        """
        return (
            cls.build_send_payload("setAttributeValue", {
                "nodeId": nodeId,
                "name": name,
                "value": value,
            }),
            None
        )

    @classmethod
    def setAttributesAsText(cls,
                            nodeId: Union['NodeId'],
                            text: Union['str'],
                            name: Optional['str'] = None,
                            ):
        """Sets attributes on element with given id. This method is useful when user edits some existing
attribute value and types in several attribute name/value pairs.
        :param nodeId: Id of the element to set attributes for.
        :type nodeId: NodeId
        :param text: Text with a number of attributes. Will parse this text using HTML parser.
        :type text: str
        :param name: Attribute name to replace with new attributes derived from text in case text parsed
successfully.
        :type name: str
        """
        return (
            cls.build_send_payload("setAttributesAsText", {
                "nodeId": nodeId,
                "text": text,
                "name": name,
            }),
            None
        )

    @classmethod
    def setFileInputFiles(cls,
                          files: Union['[]'],
                          nodeId: Optional['NodeId'] = None,
                          backendNodeId: Optional['BackendNodeId'] = None,
                          objectId: Optional['Runtime.RemoteObjectId'] = None,
                          ):
        """Sets files for the given file input element.
        :param files: Array of file paths to set.
        :type files: []
        :param nodeId: Identifier of the node.
        :type nodeId: NodeId
        :param backendNodeId: Identifier of the backend node.
        :type backendNodeId: BackendNodeId
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("setFileInputFiles", {
                "files": files,
                "nodeId": nodeId,
                "backendNodeId": backendNodeId,
                "objectId": objectId,
            }),
            None
        )

    @classmethod
    def getFileInfo(cls,
                    objectId: Union['Runtime.RemoteObjectId'],
                    ):
        """Returns file information for the given
File wrapper.
        :param objectId: JavaScript object id of the node wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("getFileInfo", {
                "objectId": objectId,
            }),
            cls.convert_payload({
                "path": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def setInspectedNode(cls,
                         nodeId: Union['NodeId'],
                         ):
        """Enables console to refer to the node with given id via $x (see Command Line API for more details
$x functions).
        :param nodeId: DOM node id to be accessible by means of $x command line API.
        :type nodeId: NodeId
        """
        return (
            cls.build_send_payload("setInspectedNode", {
                "nodeId": nodeId,
            }),
            None
        )

    @classmethod
    def setNodeName(cls,
                    nodeId: Union['NodeId'],
                    name: Union['str'],
                    ):
        """Sets node name for a node with given id.
        :param nodeId: Id of the node to set name for.
        :type nodeId: NodeId
        :param name: New node's name.
        :type name: str
        """
        return (
            cls.build_send_payload("setNodeName", {
                "nodeId": nodeId,
                "name": name,
            }),
            cls.convert_payload({
                "nodeId": {
                    "class": NodeId,
                    "optional": False
                },
            })
        )

    @classmethod
    def setNodeValue(cls,
                     nodeId: Union['NodeId'],
                     value: Union['str'],
                     ):
        """Sets node value for a node with given id.
        :param nodeId: Id of the node to set value for.
        :type nodeId: NodeId
        :param value: New node's value.
        :type value: str
        """
        return (
            cls.build_send_payload("setNodeValue", {
                "nodeId": nodeId,
                "value": value,
            }),
            None
        )

    @classmethod
    def setOuterHTML(cls,
                     nodeId: Union['NodeId'],
                     outerHTML: Union['str'],
                     ):
        """Sets node HTML markup, returns new node id.
        :param nodeId: Id of the node to set markup for.
        :type nodeId: NodeId
        :param outerHTML: Outer HTML markup to set.
        :type outerHTML: str
        """
        return (
            cls.build_send_payload("setOuterHTML", {
                "nodeId": nodeId,
                "outerHTML": outerHTML,
            }),
            None
        )

    @classmethod
    def undo(cls):
        """Undoes the last performed action.
        """
        return (
            cls.build_send_payload("undo", {
            }),
            None
        )

    @classmethod
    def getFrameOwner(cls,
                      frameId: Union['Page.FrameId'],
                      ):
        """Returns iframe node that owns iframe with the given domain.
        :param frameId: 
        :type frameId: Page.FrameId
        """
        return (
            cls.build_send_payload("getFrameOwner", {
                "frameId": frameId,
            }),
            cls.convert_payload({
                "backendNodeId": {
                    "class": BackendNodeId,
                    "optional": False
                },
                "nodeId": {
                    "class": NodeId,
                    "optional": True
                },
            })
        )



class AttributeModifiedEvent(BaseEvent):

    js_name = 'Dom.attributeModified'
    hashable = ['nodeId']
    is_hashable = True

    def __init__(self,
                 nodeId: Union['NodeId', dict],
                 name: Union['str', dict],
                 value: Union['str', dict],
                 ):
        if isinstance(nodeId, dict):
            nodeId = NodeId(**nodeId)
        self.nodeId = nodeId
        if isinstance(name, dict):
            name = str(**name)
        self.name = name
        if isinstance(value, dict):
            value = str(**value)
        self.value = value

    @classmethod
    def build_hash(cls, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class AttributeRemovedEvent(BaseEvent):

    js_name = 'Dom.attributeRemoved'
    hashable = ['nodeId']
    is_hashable = True

    def __init__(self,
                 nodeId: Union['NodeId', dict],
                 name: Union['str', dict],
                 ):
        if isinstance(nodeId, dict):
            nodeId = NodeId(**nodeId)
        self.nodeId = nodeId
        if isinstance(name, dict):
            name = str(**name)
        self.name = name

    @classmethod
    def build_hash(cls, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class CharacterDataModifiedEvent(BaseEvent):

    js_name = 'Dom.characterDataModified'
    hashable = ['nodeId']
    is_hashable = True

    def __init__(self,
                 nodeId: Union['NodeId', dict],
                 characterData: Union['str', dict],
                 ):
        if isinstance(nodeId, dict):
            nodeId = NodeId(**nodeId)
        self.nodeId = nodeId
        if isinstance(characterData, dict):
            characterData = str(**characterData)
        self.characterData = characterData

    @classmethod
    def build_hash(cls, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ChildNodeCountUpdatedEvent(BaseEvent):

    js_name = 'Dom.childNodeCountUpdated'
    hashable = ['nodeId']
    is_hashable = True

    def __init__(self,
                 nodeId: Union['NodeId', dict],
                 childNodeCount: Union['int', dict],
                 ):
        if isinstance(nodeId, dict):
            nodeId = NodeId(**nodeId)
        self.nodeId = nodeId
        if isinstance(childNodeCount, dict):
            childNodeCount = int(**childNodeCount)
        self.childNodeCount = childNodeCount

    @classmethod
    def build_hash(cls, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ChildNodeInsertedEvent(BaseEvent):

    js_name = 'Dom.childNodeInserted'
    hashable = ['previousNodeId', 'parentNodeId']
    is_hashable = True

    def __init__(self,
                 parentNodeId: Union['NodeId', dict],
                 previousNodeId: Union['NodeId', dict],
                 node: Union['Node', dict],
                 ):
        if isinstance(parentNodeId, dict):
            parentNodeId = NodeId(**parentNodeId)
        self.parentNodeId = parentNodeId
        if isinstance(previousNodeId, dict):
            previousNodeId = NodeId(**previousNodeId)
        self.previousNodeId = previousNodeId
        if isinstance(node, dict):
            node = Node(**node)
        self.node = node

    @classmethod
    def build_hash(cls, previousNodeId, parentNodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ChildNodeRemovedEvent(BaseEvent):

    js_name = 'Dom.childNodeRemoved'
    hashable = ['parentNodeId', 'nodeId']
    is_hashable = True

    def __init__(self,
                 parentNodeId: Union['NodeId', dict],
                 nodeId: Union['NodeId', dict],
                 ):
        if isinstance(parentNodeId, dict):
            parentNodeId = NodeId(**parentNodeId)
        self.parentNodeId = parentNodeId
        if isinstance(nodeId, dict):
            nodeId = NodeId(**nodeId)
        self.nodeId = nodeId

    @classmethod
    def build_hash(cls, parentNodeId, nodeId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DistributedNodesUpdatedEvent(BaseEvent):

    js_name = 'Dom.distributedNodesUpdated'
    hashable = ['insertionPointId']
    is_hashable = True

    def __init__(self,
                 insertionPointId: Union['NodeId', dict],
                 distributedNodes: Union['[BackendNode]', dict],
                 ):
        if isinstance(insertionPointId, dict):
            insertionPointId = NodeId(**insertionPointId)
        self.insertionPointId = insertionPointId
        if isinstance(distributedNodes, dict):
            distributedNodes = [BackendNode](**distributedNodes)
        self.distributedNodes = distributedNodes

    @classmethod
    def build_hash(cls, insertionPointId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class DocumentUpdatedEvent(BaseEvent):

    js_name = 'Dom.documentUpdated'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class InlineStyleInvalidatedEvent(BaseEvent):

    js_name = 'Dom.inlineStyleInvalidated'
    hashable = ['nodeIds']
    is_hashable = True

    def __init__(self,
                 nodeIds: Union['[NodeId]', dict],
                 ):
        if isinstance(nodeIds, dict):
            nodeIds = [NodeId](**nodeIds)
        self.nodeIds = nodeIds

    @classmethod
    def build_hash(cls, nodeIds):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class PseudoElementAddedEvent(BaseEvent):

    js_name = 'Dom.pseudoElementAdded'
    hashable = ['parentId']
    is_hashable = True

    def __init__(self,
                 parentId: Union['NodeId', dict],
                 pseudoElement: Union['Node', dict],
                 ):
        if isinstance(parentId, dict):
            parentId = NodeId(**parentId)
        self.parentId = parentId
        if isinstance(pseudoElement, dict):
            pseudoElement = Node(**pseudoElement)
        self.pseudoElement = pseudoElement

    @classmethod
    def build_hash(cls, parentId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class PseudoElementRemovedEvent(BaseEvent):

    js_name = 'Dom.pseudoElementRemoved'
    hashable = ['pseudoElementId', 'parentId']
    is_hashable = True

    def __init__(self,
                 parentId: Union['NodeId', dict],
                 pseudoElementId: Union['NodeId', dict],
                 ):
        if isinstance(parentId, dict):
            parentId = NodeId(**parentId)
        self.parentId = parentId
        if isinstance(pseudoElementId, dict):
            pseudoElementId = NodeId(**pseudoElementId)
        self.pseudoElementId = pseudoElementId

    @classmethod
    def build_hash(cls, pseudoElementId, parentId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class SetChildNodesEvent(BaseEvent):

    js_name = 'Dom.setChildNodes'
    hashable = ['parentId']
    is_hashable = True

    def __init__(self,
                 parentId: Union['NodeId', dict],
                 nodes: Union['[Node]', dict],
                 ):
        if isinstance(parentId, dict):
            parentId = NodeId(**parentId)
        self.parentId = parentId
        if isinstance(nodes, dict):
            nodes = [Node](**nodes)
        self.nodes = nodes

    @classmethod
    def build_hash(cls, parentId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ShadowRootPoppedEvent(BaseEvent):

    js_name = 'Dom.shadowRootPopped'
    hashable = ['rootId', 'hostId']
    is_hashable = True

    def __init__(self,
                 hostId: Union['NodeId', dict],
                 rootId: Union['NodeId', dict],
                 ):
        if isinstance(hostId, dict):
            hostId = NodeId(**hostId)
        self.hostId = hostId
        if isinstance(rootId, dict):
            rootId = NodeId(**rootId)
        self.rootId = rootId

    @classmethod
    def build_hash(cls, rootId, hostId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ShadowRootPushedEvent(BaseEvent):

    js_name = 'Dom.shadowRootPushed'
    hashable = ['hostId']
    is_hashable = True

    def __init__(self,
                 hostId: Union['NodeId', dict],
                 root: Union['Node', dict],
                 ):
        if isinstance(hostId, dict):
            hostId = NodeId(**hostId)
        self.hostId = hostId
        if isinstance(root, dict):
            root = Node(**root)
        self.root = root

    @classmethod
    def build_hash(cls, hostId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# https://stackoverflow.com/questions/30155138/how-can-i-write-asyncio-coroutines-that-optionally-act-as-regular-functions
import asyncio


class SyncAdder(type):
    """ A metaclass which adds synchronous version of coroutines.

    This metaclass finds all coroutine functions defined on a class
    and adds a synchronous version with a '_s' suffix appended to the
    original function name.

    """
    def __new__(cls, clsname, bases, dct, **kwargs):
        new_dct = {}
        for name,val in dct.items():
            # Make a sync version of all coroutine functions
            if asyncio.iscoroutinefunction(val):
                meth = cls.sync_maker(name)
                syncname = '{}_s'.format(name)
                meth.__name__ = syncname
                meth.__qualname__ = '{}.{}'.format(clsname, syncname)
                new_dct[syncname] = meth
        dct.update(new_dct)
        return super().__new__(cls, clsname, bases, dct)

    @staticmethod
    def sync_maker(func):
        def sync_func(self, *args, **kwargs):
            meth = getattr(self, func)
            return asyncio.get_event_loop().run_until_complete(meth(*args, **kwargs))
        return sync_func
